import numpy as np
import scipy as scp
from .custom_types import Constraint, ConstType, LambdaTerm, PenType
from .matrix_solvers import translate_sparse
from collections.abc import Callable
import sys

################################## Penalty functions ################################## # noqa: E266


def sort_penalties(penalties: list[LambdaTerm]) -> list[LambdaTerm]:
    """Sorts penalties by ``start_index`` in ascending order.

    :param penalties: A list of term-specific penalties.
    :type penalties: list[LambdaTerm]
    :return: A list of term-specific penalties, sorted by start index.
    :rtype: list[LambdaTerm]
    """

    idx = [pen.start_index for pen in penalties]
    sort_idx = np.argsort(idx)
    return [penalties[sidx] for sidx in sort_idx]


def create_id_dict(penalties: list[LambdaTerm]) -> dict | None:
    """Identifies penalties that should share a lambda parameter and fills a dictionary holding
    penalty indices for each shared penalty (id).

    :param penalties: A list of term-specific penalties, some of which might have a ``id`` flag.
    :type penalties: list[LambdaTerm]
    :return: A dictionary. Keys are ids shared by two or more penalties. Values are lists holding
        corresponding penalty indices. If no id exists that is shared by two or more penalties,
        None is returned instead.
    :rtype: dict|None
    """
    id_dict = {}

    for peni, pen in enumerate(penalties):
        if pen.id is not None:
            if pen.id in id_dict:
                id_dict[pen.id].append(peni)
            else:
                id_dict[pen.id] = [peni]

    if len(id_dict) > 0 and max([len(v) for v in id_dict.values()]) > 1:
        return id_dict

    return None


def combine_shared_penalties(penalties: list[LambdaTerm]) -> list[LambdaTerm]:
    """Identifies penalties that should share a lambda parameter and merges them into a single
    :class:`LambdaTerm`.

    :param penalties: A list of term-specific penalties, some of which might have a ``id`` flag.
    :type penalties: list[LambdaTerm]
    :return: A list of penalty matrices in which individual penalties with a shared ``id`` flag
        have been combined into a single :class:`LambdaTerm`.
    :rtype: list[LambdaTerm]
    """

    id_dict = create_id_dict(penalties)

    if id_dict is None:
        return penalties
    else:
        merged_penalties = [
            pen
            for pen in penalties
            if (pen.id not in id_dict) or (len(id_dict[pen.id]) <= 1)
        ]

        for id in id_dict.keys():
            m_pens = [pen for pen in penalties if pen.id == id]
            S_J_emb = m_pens[0].S_J_emb
            D_J_emb = m_pens[0].D_J_emb

            for peni in range(1, len(m_pens)):
                S_J_emb += m_pens[peni].S_J_emb
                D_J_emb += m_pens[peni].D_J_emb

            m_pen = LambdaTerm(
                S_J=S_J_emb[
                    m_pens[0].start_index : (  # noqa: E203
                        m_pens[-1].start_index
                        + (m_pens[-1].S_J.shape[1] * m_pens[-1].rep_sj)
                    ),
                    m_pens[0].start_index : (  # noqa: E203
                        m_pens[-1].start_index
                        + (m_pens[-1].S_J.shape[1] * m_pens[-1].rep_sj)
                    ),
                ],
                S_J_emb=S_J_emb,
                D_J_emb=D_J_emb,
                lam=m_pens[0].lam,
                rep_sj=1,
                start_index=m_pens[0].start_index,
                type=PenType.SHARED,
                rank=np.sum([pen.rank for pen in m_pens]),
                term=[pen.term for pen in m_pens],
                dist_param=[pen.dist_param for pen in m_pens],
                S_Js=[pen.S_J for pen in m_pens],
                S_J_embs=[pen.S_J_emb for pen in m_pens],
                D_J_embs=[pen.D_J_emb for pen in m_pens],
                rep_sjs=[pen.rep_sj for pen in m_pens],
                start_indices=[pen.start_index for pen in m_pens],
                types=[pen.type for pen in m_pens],
                ranks=[pen.rank for pen in m_pens],
                id=id,
            )

            merged_penalties.append(m_pen)

        return merged_penalties


def split_shared_penalties(merged_penalties: list[LambdaTerm]) -> list[LambdaTerm]:
    """Identifies penalties that share a lambda parameter and splits them into individual
    :class:`LambdaTerm`s - all having the same :math:`\\lambda` value.

    Basically inverts what is achieved by the :func:`combine_shared_penalties` function.

    :param merged_penalties:  list of penalty matrices in which individual penalties with a shared
        ``id`` flag have been combined into a single :class:`LambdaTerm`.
    :type merged_penalties: list[LambdaTerm]
    :return: A list of penalty matrices in which merged penalties have been split into individual
        penalties again.
    :rtype: list[LambdaTerm]
    """

    merged_idx = []
    for peni, pen in enumerate(merged_penalties):
        if isinstance(pen.S_Js, list):
            merged_idx.append(peni)

    penalties = [
        merged_penalties[peni]
        for peni in range(len(merged_penalties))
        if peni not in merged_idx
    ]

    if len(merged_idx) == 0:
        return penalties

    else:
        for idx in merged_idx:

            merged_pen = merged_penalties[idx]

            for m in range(len(merged_pen.S_Js)):

                penalties.append(
                    LambdaTerm(
                        S_J=merged_pen.S_Js[m],
                        S_J_emb=merged_pen.S_J_embs[m],
                        D_J_emb=merged_pen.D_J_embs[m],
                        lam=merged_pen.lam,
                        rep_sj=merged_pen.rep_sjs[m],
                        start_index=merged_pen.start_indices[m],
                        type=merged_pen.types[m],
                        rank=merged_pen.ranks[m],
                        term=merged_pen.term[m],
                        dist_param=merged_pen.dist_param[m],
                        id=merged_pen.id,
                    )
                )

        # Sort back by start indices
        penalties = sort_penalties(penalties)
        return penalties


def adjust_pen_drop(
    dat: list[float], rows: list[int], cols: list[int], drop: list[int], offset: int = 0
) -> tuple[list[float], list[int], list[int], int]:
    """Adjusts penalty matrix (represented via ``dat``, ``rows``, and ``cols``) by dropping rows
    and columns indicated by ``drop``.

    Optionally, ``offset`` is added to the elements in ``rows`` and ``cols``, which is useful when
    indices in ``drop`` do not start at zero.

    :param dat: List of elements in penalty matrix.
    :type dat: [float]
    :param rows: List of row indices of penalty matrix.
    :type rows: [int]
    :param cols: List of column indices of penalty matrix.
    :type cols: [int]
    :param drop: Rows and columns to drop from penalty matrix. Might actually contain indices
        corresponding to ``rows + offset`` and ``cols + offset``, which can be corrected for via
        the ``offset`` argument.
    :type drop: [int]
    :param offset: An optional offset to add to ``rows`` and ``cols`` to adjust for the indexing in
        ``drop``, defaults to 0
    :type offset: int, optional
    :return: A tuple with 4 elements: the data, rows, and cols of the adjusted penalty matrix
        excluding dropped elements and the number of excluded elements.
    :rtype: tuple[list[float],list[int],list[int],int]
    """
    rows = np.array(rows)
    cols = np.array(cols)
    dat = np.array(dat)

    drop_idx = np.isin(drop, cols + offset)
    dropped = np.sum(drop_idx)
    drop = np.array(drop)[drop_idx]

    keep_col = ~np.isin(cols + offset, drop)
    keep_row = ~np.isin(rows + offset, drop)

    keep = keep_col & keep_row

    # Now adjust cols & rows
    rows_realign = np.zeros_like(rows)
    cols_realign = np.zeros_like(cols)

    rows_realign[:] = rows
    cols_realign[:] = cols

    for d in drop:
        rows_realign[rows + offset > d] -= 1
        cols_realign[cols + offset > d] -= 1

    # Now return
    return list(dat[keep]), list(rows_realign[keep]), list(cols_realign[keep]), dropped


def embed_in_S_sparse(
    pen_data: list[float],
    pen_rows: list[int],
    pen_cols: list[int],
    S_emb: scp.sparse.csc_array | None,
    S_col: int,
    SJ_col: int,
    cIndex: int,
) -> tuple[scp.sparse.csc_array, int]:
    """Embed a term-specific penalty matrix ``SJ`` (provided as three lists: ``pen_data``,
    ``pen_rows`` and ``pen_cols``) into the total penalty matrix ``S_emb`` (see Wood, 2017)

    :param pen_data: Data of ``SJ``
    :type pen_data: list[float]
    :param pen_rows: Row indices of ``SJ``
    :type pen_rows: list[int]
    :param pen_cols: Column indices of ``SJ``
    :type pen_cols: list[int]
    :param S_emb: Total penalty matrix or ``None`` in case ``S_emb`` will be initialized by the
        function.
    :type S_emb: scp.sparse.csc_array | None
    :param S_col: Columns of total penalty matrix
    :type S_col: int
    :param SJ_col: Columns of ``SJ``
    :type SJ_col: int
    :param cIndex: Current row and column index indicating the top left cell of the
        (``SJ_col`` * ``SJ_col``) block ``SJ`` should take up in ``S_emb``
    :type cIndex: int
    :return: ``S_emb`` with ``SJ`` embedded, the updated ``cIndex`` (i.e., ``cIndex + SJ_col``)
    :rtype: tuple[scp.sparse.csc_array,int]
    """

    embedding = np.array(pen_data)
    r_embedding = np.array(pen_rows) + cIndex
    c_embedding = np.array(pen_cols) + cIndex

    if S_emb is None:
        S_emb = scp.sparse.csc_array(
            (embedding, (r_embedding, c_embedding)), shape=(S_col, S_col)
        )
    else:
        S_emb += scp.sparse.csc_array(
            (embedding, (r_embedding, c_embedding)), shape=(S_col, S_col)
        )

    return S_emb, cIndex + SJ_col


def embed_in_Sj_sparse(
    pen_data: list[float],
    pen_rows: list[int],
    pen_cols: list[int],
    Sj: scp.sparse.csc_array | None,
    SJ_col: int,
) -> scp.sparse.csc_array:
    """Parameterize a term-specific penalty matrix ``SJ`` (provided as three lists: ``pen_data``,
    ``pen_rows`` and ``pen_cols``).

    :param pen_data: Data of ``SJ``
    :type pen_data: list[float]
    :param pen_rows: Row indices of ``SJ``
    :type pen_rows: list[int]
    :param pen_cols: Column indices of ``SJ``
    :type pen_cols: list[int]
    :param Sj: A sparse matrix or ``None``. In the latter case, ``SJ`` is simply initialized by the
        function. If not, then the function returns ``SJ + Sj``. The latter is useful if a term
        penalty is a sum of individual penalty matrices.
    :type Sj: scp.sparse.csc_array | None
    :param SJ_col: Columns of ``SJ``
    :type SJ_col: int
    :return: ``SJ`` which might actually be ``SJ + Sj``.
    :rtype: scp.sparse.csc_array
    """
    embedding = np.array(pen_data)

    if Sj is None:
        Sj = scp.sparse.csc_array(
            (embedding, (pen_rows, pen_cols)), shape=(SJ_col, SJ_col)
        )
    else:
        Sj += scp.sparse.csc_array(
            (embedding, (pen_rows, pen_cols)), shape=(SJ_col, SJ_col)
        )

    return Sj


def embed_shared_penalties(
    shared_penalties: list[list[LambdaTerm]], formulas: list, extra_coef: int
) -> list[LambdaTerm]:
    """Embed penalties from individual formulas into overall penalties for GAMMLSS/GSMM models.

    :param shared_penalties: Nested list, with the inner one containing the penalties associated
        with an individual formula in ``formulas``.
    :type shared_penalties: list[list[LambdaTerm]]
    :param formulas: List of :class:`mssm.src.python.formula.Formula` objects
    :type formulas: list
    :param extra_coef: Number of extra coefficients required by the model's family. Will result in
        the shared penalties being padded by an extra block of ``extra_coef`` zeroes.
    :type extra_coef: int
    :return: A list of the embedded penalties required by a GAMMLSS or GSMM model.
    :rtype: list[LambdaTerm]
    """
    # Assign original formula index to each penalty
    for fi in range(len(shared_penalties)):
        for lterm in shared_penalties[fi]:
            lterm.dist_param = fi

    for fi, form in enumerate(formulas):
        for ofi, other_form in enumerate(formulas):
            if fi == ofi:
                continue

            if ofi < fi:
                for lterm in shared_penalties[fi]:
                    lterm.S_J_emb = scp.sparse.vstack(
                        [
                            scp.sparse.csc_array(
                                (other_form.n_coef, lterm.S_J_emb.shape[1])
                            ),
                            lterm.S_J_emb,
                        ]
                    ).tocsc()
                    lterm.D_J_emb = scp.sparse.vstack(
                        [
                            scp.sparse.csc_array(
                                (other_form.n_coef, lterm.S_J_emb.shape[1])
                            ),
                            lterm.D_J_emb,
                        ]
                    ).tocsc()

                    lterm.S_J_emb = scp.sparse.hstack(
                        [
                            scp.sparse.csc_array(
                                (lterm.S_J_emb.shape[0], other_form.n_coef)
                            ),
                            lterm.S_J_emb,
                        ]
                    ).tocsc()

                    lterm.D_J_emb = scp.sparse.hstack(
                        [
                            scp.sparse.csc_array(
                                (lterm.S_J_emb.shape[0], other_form.n_coef)
                            ),
                            lterm.D_J_emb,
                        ]
                    ).tocsc()

                    lterm.start_index += other_form.n_coef

            elif ofi > fi:
                for lterm in shared_penalties[fi]:
                    lterm.S_J_emb = scp.sparse.vstack(
                        [
                            lterm.S_J_emb,
                            scp.sparse.csc_array(
                                (other_form.n_coef, lterm.S_J_emb.shape[1])
                            ),
                        ]
                    ).tocsc()

                    lterm.D_J_emb = scp.sparse.vstack(
                        [
                            lterm.D_J_emb,
                            scp.sparse.csc_array(
                                (other_form.n_coef, lterm.S_J_emb.shape[1])
                            ),
                        ]
                    ).tocsc()

                    lterm.S_J_emb = scp.sparse.hstack(
                        [
                            lterm.S_J_emb,
                            scp.sparse.csc_array(
                                (lterm.S_J_emb.shape[0], other_form.n_coef)
                            ),
                        ]
                    ).tocsc()

                    lterm.D_J_emb = scp.sparse.hstack(
                        [
                            lterm.D_J_emb,
                            scp.sparse.csc_array(
                                (lterm.S_J_emb.shape[0], other_form.n_coef)
                            ),
                        ]
                    ).tocsc()

    if extra_coef is not None:
        for fi, form in enumerate(formulas):
            for lterm in shared_penalties[fi]:
                lterm.S_J_emb = scp.sparse.vstack(
                    [
                        lterm.S_J_emb,
                        scp.sparse.csc_array((extra_coef, lterm.S_J_emb.shape[1])),
                    ]
                ).tocsc()

                lterm.D_J_emb = scp.sparse.vstack(
                    [
                        lterm.D_J_emb,
                        scp.sparse.csc_array((extra_coef, lterm.S_J_emb.shape[1])),
                    ]
                ).tocsc()

                lterm.S_J_emb = scp.sparse.hstack(
                    [
                        lterm.S_J_emb,
                        scp.sparse.csc_array((lterm.S_J_emb.shape[0], extra_coef)),
                    ]
                ).tocsc()

                lterm.D_J_emb = scp.sparse.hstack(
                    [
                        lterm.D_J_emb,
                        scp.sparse.csc_array((lterm.S_J_emb.shape[0], extra_coef)),
                    ]
                ).tocsc()

    return shared_penalties


class Penalty:
    """Penalty base-class. Generates penalty matrices for smooth terms.

    :param pen_type: Type of the penalty matrix
    :type pen_type: PenType
    :ivar PenType pen_type: Type of the penalty matrix passed to the init method.
    """

    def __init__(self, pen_type: PenType) -> None:
        self.type = pen_type

    def constructor(
        self, n: int, constraint: Constraint | None, *args, **kwargs
    ) -> tuple[
        list[float], list[int], list[int], list[float], list[int], list[int], int
    ]:
        """Creates penalty matrix + root of the penalty and returns both in list form (data, row
        indices, col indices).

        :param n: Dimension of square penalty matrix
        :type n: int
        :param constraint: Any contraint to absorb by the penalty or None if no constraint is
            required
        :type constraint: Constraint | None
        :return: penalty data, penalty row indices, penalty column indices, root of penalty data,
            root of penalty row indices, root of penalty column indices, rank of penalty
        :rtype: tuple[list[float], list[int], list[int], list[float], list[int], list[int], int]
        """
        pass


class DifferencePenalty(Penalty):
    """Difference Penalty class. Generates penalty matrices for smooth terms.

    :ivar PenType.DIFFERENCE pen_type: Type of the penalty matrix.
    """

    def __init__(self):
        super().__init__(PenType.DIFFERENCE)

    def constructor(
        self, n: int, constraint: Constraint | None, m: int = 2
    ) -> tuple[
        list[float], list[int], list[int], list[float], list[int], list[int], int
    ]:
        """Creates difference (order=m) n*n penalty matrix + root of the penalty. Based on code in
        Eilers & Marx (1996) and Wood (2017).

        References:
           - Eilers, P. H. C., & Marx, B. D. (1996). Flexible smoothing with B-splines and \
            penalties. Statistical Science, 11(2), 89–121. https://doi.org/10.1214/ss/1038425655
           - Wood, S. N. (2017). Generalized Additive Models: An Introduction with R, Second \
            Edition (2nd ed.).

        :param n: Dimension of square penalty matrix
        :type n: int
        :param constraint: Any contraint to absorb by the penalty or None if no constraint is
            required
        :type constraint: Constraint|None
        :param m: Differencing order to apply to the identity matrix to get the penalty (this will
            also be the dimension of the penalty's Kernel), defaults to 2
        :type m: int, optional
        :return: penalty data,penalty row indices,penalty column indices,root of penalty data,root
            of penalty row indices,root of penalty column indices,rank of penalty
        :rtype: tuple[list[float],list[int],list[int],list[float],list[int],list[int],int]
        """
        D = np.diff(np.identity(n), m)
        S = D @ D.T
        # Eilers & Marx (1996): P-spline penalties consider m-degree polynomial as
        # smooth, i.e., un-penalized!
        rank = n - m

        # Absorb any identifiability constraints
        if constraint is not None:
            Z = constraint.Z
            if constraint.type == ConstType.QR:
                S = Z.T @ S @ Z
                D = Z.T @ D
            elif constraint.type == ConstType.DROP:
                S = np.delete(np.delete(S, Z, axis=1), Z, axis=0)
                D = np.delete(D, Z, axis=0)
            elif constraint.type == ConstType.DIFF:
                if m == 0:
                    raise ValueError(
                        (
                            "When using ConstType.DIFF, for a term with a difference penalty, `m` "
                            "must be greater than zero!"
                        )
                    )
                # Correct for column differencing applied to X! See smoothCon help for
                # mgcv (Wood, 2017)
                D = np.diff(
                    np.concatenate(
                        (D[Z : D.shape[0], :], D[:Z, :]), axis=0  # noqa: E203
                    ),
                    axis=0,
                )
                D = np.concatenate(
                    (D[D.shape[0] - Z :, :], D[: D.shape[0] - Z, :]),  # noqa: E203
                    axis=0,
                )
                S = D @ D.T

            if m == 0:
                # Re-compute root
                eig, U = scp.linalg.eigh(S)
                D = U @ np.diag(
                    [
                        (
                            np.power(e, 0.5)
                            if e > np.power(sys.float_info.epsilon, 0.7)
                            else 0
                        )
                        for e in eig
                    ]
                )

        S = scp.sparse.csc_array(S)
        D = scp.sparse.csc_array(D)

        # Data in S and D is in canonical format, for competability this is translated to
        # data, rows, columns
        pen_data, pen_rows, pen_cols = translate_sparse(S)
        chol_data, chol_rows, chol_cols = translate_sparse(D)

        return pen_data, pen_rows, pen_cols, chol_data, chol_rows, chol_cols, rank


class IdentityPenalty(Penalty):
    """Difference Penalty class. Generates penalty matrices for smooth terms and random terms.

    :param pen_type: Type of the penalty matrix
    :type pen_type: PenType
    :ivar PenType pen_type: Type of the penalty matrix passed to init method.
    """

    def __init__(self, pen_type: PenType):
        if pen_type not in [PenType.IDENTITY, PenType.DISTANCE]:
            raise ValueError(
                f"pen_type must be PenType.IDENTITY or PenType.DISTANCE, but is {pen_type}"
            )
        super().__init__(pen_type)

    def constructor(
        self, n: int, constraint: Constraint | None, f: Callable | None = None
    ) -> tuple[
        list[float], list[int], list[int], list[float], list[int], list[int], int
    ]:
        """Creates identity matrix penalty + root in case ``f is None``.

        **Note**: This penalty never absorbs marginal constraints. It always returns an identity
        matrix but just decreases ``n`` by 1 if ``constraint is not None``
        to ensure that the returned penalty matrix is of suitable dimensions.

        :param n: Dimension of square penalty matrix
        :type n: int
        :param constraint: Any contraint to absorb by the penalty or None if no constraint is
            required
        :type constraint: Constraint|None
        :param f: Any kind of function to apply to the diagonal elements of the penalty, defaults
            to None
        :type f: Callable|None, optional
        :return: penalty data, penalty row indices, penalty column indices, root of penalty data,
            root of penalty row indices, root of penalty column indices, rank of penalty
        :rtype: tuple[list[float],list[int],list[int],list[float],list[int],list[int],int]
        """
        # Can be used to create event-distance weighted penalty matrices for deconvolving sms GAMMs
        if constraint is not None:
            n -= 1

        elements = [0.0 for _ in range(n)]
        idx = [0.0 for _ in range(n)]

        for i in range(n):
            if f is None:
                elements[i] = 1.0
            else:
                elements[i] = f(i + 1)
            idx[i] = i

        return (
            elements,
            idx,
            idx,
            elements,
            idx,
            idx,
            n,
        )  # I' @ I = I; also identity is full rank


def TP_pen(
    S_j: scp.sparse.csc_array,
    D_j: scp.sparse.csc_array,
    j: int,
    ks: list[int],
    constraint: Constraint | None,
    scale_pen: bool,
) -> tuple[list[float], list[int], list[int], list[float], list[int], list[int], int]:
    """Computes a tensor smooth penalty + root as defined in section 5.6 of Wood (2017) based on
    marginal penalty matrix ``S_j``.

    References:
     - Wood, S. N. (2017). Generalized Additive Models: An Introduction with R, Second Edition \
        (2nd ed.).

    :param S_j: Marginal penalty matrix
    :type S_j: scp.sparse.csc_array
    :param D_j: Root of marginal penalty matrix
    :type D_j: scp.sparse.csc_array
    :param j: Index for current marginal
    :type j: int
    :param ks: List of number of basis functions of all marginals
    :type ks: list[int]
    :param constraint: Any constraint to absorb by the final penalty or None if no constraint is
        required
    :type constraint: Constraint | None
    :return: penalty data, penalty row indices, penalty column indices, root of penalty data,
        root of penalty row indices, root of penalty column indices, rank of penalty
    :rtype: tuple[list[float],list[int],list[int],list[float],list[int],list[int],int]
    """
    # Tensor smooth penalty - not including the reparameterization of Wood (2017) 5.6.2
    # but reflecting Eilers & Marx (2003) instead

    # Apply scaling as done by mgcv to improve numerical stability of tensor smooth terms.
    # see: https://github.com/cran/mgcv/blob/master/R/smooth.r#L824
    if scale_pen:
        eig, _ = scp.linalg.eigh(S_j.toarray())
        S_j /= eig[-1]
        D_j /= np.sqrt(eig[-1])

    if j == 0:
        S_TP = S_j
        D_TP = D_j
    else:
        S_TP = scp.sparse.identity(ks[0])
        D_TP = scp.sparse.identity(ks[0])
        # m_rank *= ks[0] # Modify rank of marginal - identities are full-rank.

    for i in range(1, len(ks)):
        if j == i:
            S_TP = scp.sparse.kron(S_TP, S_j, format="csc")
            D_TP = scp.sparse.kron(D_TP, D_j, format="csc")
        else:
            S_TP = scp.sparse.kron(S_TP, scp.sparse.identity(ks[i]), format="csc")
            D_TP = scp.sparse.kron(D_TP, scp.sparse.identity(ks[i]), format="csc")
            # m_rank *= ks[i]

    if constraint is not None:
        Z = constraint.Z
        if constraint.type == ConstType.QR:
            S_TP = Z.T @ S_TP @ Z
            D_TP = Z.T @ D_TP
            S_TP = scp.sparse.csc_array(S_TP)
            D_TP = scp.sparse.csc_array(D_TP)
        elif constraint.type == ConstType.DROP:
            S_TP = scp.sparse.csc_array(
                np.delete(np.delete(S_TP.toarray(), Z, axis=1), Z, axis=0)
            )
            D_TP = scp.sparse.csc_array(np.delete(D_TP.toarray(), Z, axis=0))
        elif constraint.type == ConstType.DIFF:
            D_TP = D_TP.toarray()
            # Correct for column differencing applied to X! See smoothCon help for mgcv (Wood, 2017)
            D_TP = np.diff(
                np.concatenate(
                    (D_TP[Z : D_TP.shape[0], :], D_TP[:Z, :]), axis=0  # noqa: E203
                ),
                axis=0,
            )
            D_TP = np.concatenate(
                (
                    D_TP[D_TP.shape[0] - Z :, :],  # noqa: E203
                    D_TP[: D_TP.shape[0] - Z, :],
                ),
                axis=0,
            )
            S_TP = D_TP @ D_TP.T
            S_TP = scp.sparse.csc_array(S_TP)
            D_TP = scp.sparse.csc_array(D_TP)

        # Check for full-rank marginal -> need new D_TP
        _, _, chol_cols = translate_sparse(D_TP)

        if (max(chol_cols) + 1) > D_TP.shape[0]:
            if constraint.type == ConstType.DIFF:
                raise ValueError(
                    (
                        "Cannot compute tensor smooth penalty for constraint.type=ConstType.DIFF. "
                        "Set a different constraint for the marginals or ensure that marginal "
                        "penalties have a non-trivial kernel!"
                    )
                )

            # Re-compute root
            eig, U = scp.linalg.eigh(S_TP.toarray())
            D_TP = scp.sparse.csc_array(
                U
                @ np.diag(
                    [
                        (
                            np.power(e, 0.5)
                            if e > np.power(sys.float_info.epsilon, 0.7)
                            else 0
                        )
                        for e in eig
                    ]
                )
            )

    pen_data, pen_rows, pen_cols = translate_sparse(S_TP)
    chol_data, chol_rows, chol_cols = translate_sparse(D_TP)
    return pen_data, pen_rows, pen_cols, chol_data, chol_rows, chol_cols



import numpy as np
import scipy as scp
import eigen_solvers
import warnings
from multiprocessing import managers, shared_memory
import multiprocessing as mp
from itertools import repeat
from .custom_types import PenType, LambdaTerm


def map_csc_to_eigen(
    X: scp.sparse.csc_array,
) -> tuple[int, int, int, np.ndarray, np.ndarray, np.ndarray]:
    """Pybind11 comes with copy overhead for sparse matrices, so instead of passing the
    sparse matrix to c++, I pass the data, indices, and indptr arrays as buffers to c++.
    see: https://pybind11.readthedocs.io/en/stable/advanced/pycpp/numpy.html.

    An Eigen mapping can then be used to refer to these, without requiring an extra copy.
    see: https://eigen.tuxfamily.org/dox/classEigen_1_1Map_3_01SparseMatrixType_01_4.html

    The mapping needs to assume compressed storage, since then we can use the indices, indptr, and
    data arrays directly for the valuepointer, innerPointer, and outerPointer fields of the
    sparse array map constructor.
    see: https://eigen.tuxfamily.org/dox/group__TutorialSparse.html (section sparse matrix format).

    I got this idea from the NumpyEigen project, which also uses such a map!
    see: https://github.com/fwilliams/numpyeigen/blob/master/src/npe_sparse_array.h#L74

    :param X: Some sparse matrix
    :type X: scp.sparse.csc_array
    :return: Number of rows in X, Number of cols in X, Number of non-zero elements in X, X.data,
        X.indptr.astype(np.int64), X.indices.astype(np.int64)
    :rtype: tuple[int,int,int,np.ndarray,np.ndarray,np.ndarray]
    """

    if X.format != "csc":
        raise TypeError(
            f"Format of sparse matrix passed to c++ MUST be 'csc' but is {X.getformat()}"
        )

    if X.has_sorted_indices is False:
        raise TypeError(
            "Indices of sparse matrix passed to c++ MUST be sorted but are not."
        )

    rows, cols = X.shape

    # Cast to int64 here, since that's what the c++ side expects to be stored in the buffers
    return (
        rows,
        cols,
        X.nnz,
        X.data,
        X.indptr.astype(np.int64),
        X.indices.astype(np.int64),
    )


def map_csr_to_eigen(
    X: scp.sparse.csr_array,
) -> tuple[int, int, int, np.ndarray, np.ndarray, np.ndarray]:
    """see: :func:`map_csc_to_eigen`

    :param X: Some sparse matrix
    :type X: scp.sparse.csr_array
    :return: Number of rows in X, Number of cols in X, Number of non-zero elements in X, X.data,
        X.indptr.astype(np.int64), X.indices.astype(np.int64)
    :rtype: tuple[int,int,int,np.ndarray,np.ndarray,np.ndarray]
    """

    if X.format != "csr":
        raise TypeError(
            f"Format of sparse matrix passed to c++ MUST be 'csr' but is {X.getformat()}"
        )

    if X.has_sorted_indices is False:
        raise TypeError(
            "Indices of sparse matrix passed to c++ MUST be sorted but are not."
        )

    rows, cols = X.shape

    # Cast to int64 here, since that's what the c++ side expects to be stored in the buffers
    return (
        rows,
        cols,
        X.nnz,
        X.data,
        X.indptr.astype(np.int64),
        X.indices.astype(np.int64),
    )


def translate_sparse(
    mat: scp.sparse.csc_array,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Translate canonical sparse csc matrix representation into data, row, col representation

    See: https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.csc_array.html#scipy.sparse.csc_array # noqa: E501

    :param mat: sparse matrix
    :type mat: scp.sparse.csc_array
    :return: data, rows, cols of sparse matrix
    :rtype: tuple[np.ndarray,np.ndarray,np.ndarray]
    """

    if mat.format != "csc":
        raise TypeError(
            (
                "Format of sparse matrix passed to be translated must be 'csc' but "
                f"is {mat.getformat()}"
            )
        )

    elements = mat.data
    idx = mat.indices
    iptr = mat.indptr

    data = []
    rows = []
    cols = []

    for ci in range(mat.shape[1]):

        c_data = elements[iptr[ci] : iptr[ci + 1]]  # noqa: E203
        c_rows = idx[iptr[ci] : iptr[ci + 1]]  # noqa: E203

        data.extend(c_data)
        rows.extend(c_rows)
        cols.extend([ci for _ in range(len(c_rows))])

    return data, rows, cols


def compute_eigen_perm(Pr: list[int] | np.ndarray) -> scp.sparse.csc_array:
    """Internal function. Computes column permutation matrix obtained from Eigen.

    :param Pr: List of column indices
    :type Pr: list[int] | np.ndarray
    :return: Permutation matrix as sparse array
    :rtype: scp.sparse.csc_array
    """

    nP = len(Pr)
    P = [1 for _ in range(nP)]
    Pc = [c for c in range(nP)]
    Perm = scp.sparse.csc_array((P, (Pr, Pc)), shape=(nP, nP))
    return Perm


def apply_eigen_perm(
    Pr: list[int], InvCholXXSP: scp.sparse.csc_array
) -> scp.sparse.csc_array:
    """Internal function. Unpivots columns of ``InvCholXXSP`` (usually the inverse of a
    Cholesky factor) and returns the unpivoted version.

    :param Pr: List of column indices
    :type Pr: list[int]
    :param InvCholXXSP: Pivoted matrix
    :type InvCholXXSP: scp.sparse.csc_array
    :return: Unpivoted matrix
    :rtype: scp.sparse.csc_array
    """
    Perm = compute_eigen_perm(Pr)
    InvCholXXS = InvCholXXSP @ Perm
    return InvCholXXS


def cpp_chol(A: scp.sparse.csc_array) -> tuple[scp.sparse.csc_array, int]:
    """Computes Cholesky of ``A``.

    :param A: Some square symmetric matrix
    :type A: scp.sparse.csc_array
    :return: Returns Cholesky and code indicating success
    :rtype: tuple[scp.sparse.csc_array,int]
    """
    return eigen_solvers.chol(*map_csc_to_eigen(A))


def cpp_cholP(A: scp.sparse.csc_array) -> tuple[scp.sparse.csc_array, list[int], int]:
    """Computes pivoted Cholesky of ``A``.

    :param A: Some square symmetric matrix
    :type A: scp.sparse.csc_array
    :return: Returns pivoted Cholesky, pivoted column order, and code indicating success
    :rtype: tuple[scp.sparse.csc_array,list[int],int]
    """
    return eigen_solvers.cholP(*map_csc_to_eigen(A))


def cpp_qr(
    A: scp.sparse.csc_array,
) -> tuple[scp.sparse.csc_array, scp.sparse.csc_array, list[int], int]:
    """Computes pivoted QR decomposition of ``A``.

    :param A: Some matrix
    :type A: scp.sparse.csc_array
    :return: Matrices Q, R, pivoted column order, and code indicating success
    :rtype: tuple[scp.sparse.csc_array,scp.sparse.csc_array,list[int],int]
    """
    return eigen_solvers.pqr(*map_csc_to_eigen(A))


def cpp_qrr(
    A: scp.sparse.csc_array,
) -> tuple[scp.sparse.csc_array, list[int], int, int]:
    """Computes pivoted QR decomposition of ``A`` and returns rank estimate

    :param A: Some matrix
    :type A: scp.sparse.csc_array
    :return: Matrices Q, R, pivoted column order, estimated rank, and code indicating success
    :rtype: tuple[scp.sparse.csc_array,list[int],int,int]
    """
    return eigen_solvers.pqrr(*map_csc_to_eigen(A))


def cpp_dqrr(A: np.ndarray) -> tuple[list[int], int]:
    """Computes pivoted QR decomposition of dense matrix ``A``.

    :param A: Some matrix
    :type A: np.ndarray
    :return: column pivot order for rank estimation, estimated rank
    :rtype: tuple[list[int],int]
    """
    return eigen_solvers.dpqrr(A)


def cpp_symqr(
    A: scp.sparse.csc_array, tol: float
) -> tuple[scp.sparse.csc_array, list[int], list[int], int, int]:
    """Computes pivoted QR decomposition of symmetric matrix ``A``.

    :param A: Some symmetric matrix
    :type A: scp.sparse.csc_array
    :param tol: tolerance for rank estimation
    :type tol: float
    :return: Matrix R, column pivot order for sparsity, column pivot order for rank estimation,
        rank estimate, code indicating success
    :rtype: tuple[scp.sparse.csc_array,list[int],list[int],int,int]
    """
    return eigen_solvers.spqr(*map_csc_to_eigen(A), tol)


def cpp_solve_qr(A: scp.sparse.csc_array) -> tuple[scp.sparse.csc_array, int, int]:
    """Solves ``A@B=I`` for ``B``, where ``A`` is sparse, square, and full rank and ``I`` is an
    identity matrix of suitable dimension via QR decomposition.

    :param A: Some sparse square matrix
    :type A: scp.sparse.csc_array
    :return: ``B`` (inverse of ``A``), estimated rank, and code indicating success
    :rtype: tuple[scp.sparse.csc_array,int,int]
    """
    return eigen_solvers.solve_pqr(*map_csc_to_eigen(A))


def cpp_solve_am(
    y: np.ndarray, X: scp.sparse.csc_array, S: scp.sparse.csc_array
) -> tuple[scp.sparse.csc_array, list[int], np.ndarray, int]:
    """Solves ``(X.T@X + S)@b = X.T@y`` for ``b`` via sparse Cholesky decomposition and computes
    inverse of pivoted Cholesky of ``X.T@X + S``.

    :param y: vector of observations
    :type y: np.ndarray
    :param X: Some rectangular sparse matrix
    :type X: scp.sparse.csc_array
    :param S: Sparse square matrix
    :type S: scp.sparse.csc_array
    :return: Inverse of pivoted Cholesky of ``X.T@X + S``, column pivot indices in a list, ``b``,
        and code indicating success
    :rtype: tuple[scp.sparse.csc_array,list[int],np.ndarray,int]
    """
    return eigen_solvers.solve_am(y, *map_csc_to_eigen(X), *map_csc_to_eigen(S))


def cpp_solve_coef(
    y: np.ndarray, X: scp.sparse.csc_array, S: scp.sparse.csc_array
) -> tuple[scp.sparse.csc_array, list[int], np.ndarray, int]:
    """Solves ``(X.T@X + S)@b = X.T@y`` for ``b`` via sparse Cholesky decomposition.

    :param y: vector of observations
    :type y: np.ndarray
    :param X: Some rectangular sparse matrix
    :type X: scp.sparse.csc_array
    :param S: Sparse square matrix
    :type S: scp.sparse.csc_array
    :return: Pivoted Cholesky of ``X.T@X + S``, column pivot indices in a list, ``b``, and code
        indicating success
    :rtype: tuple[scp.sparse.csc_array,list[int],np.ndarray,int]
    """
    return eigen_solvers.solve_coef(y, *map_csc_to_eigen(X), *map_csc_to_eigen(S))


def cpp_solve_coef_pqr(
    y: np.ndarray, X: scp.sparse.csc_array, E: scp.sparse.csc_array
) -> tuple[scp.sparse.csc_array, list[int], list[int], np.ndarray, int, int]:
    """Solves ``(X.T@X + S)@b = X.T@y`` for ``b`` via sparse QR decomposition, where ``E.T@E=S``.

    **Does not form ``X.T@X + S`` for solve**. Potentially pivots twice - once for sparsity (always)
    and then once more whenever algorithm detects a diagonal element that is too small.

    Examples::

       # Solve
       RP,Pr1,Pr2,coef,rank,code = cpp_solve_coef_pqr(yb,Xb,S_root.T.tocsc())

       # Need to get overall pivot...
       P1 = compute_eigen_perm(Pr1)
       P2 = compute_eigen_perm(Pr2)
       P = P2.T@P1.T

       # Need to insert zeroes in case of rank deficiency - first insert nans to that we
       # can then easily find dropped coefs.
       if rank < S_emb.shape[1]:
          coef = np.concatenate((coef,[np.nan for _ in range(S_emb.shape[1]-rank)]))

       # Can now unpivot coef
       coef = coef @ P

       # And identify which coef was dropped
       idx = np.arange(len(coef))
       drop = idx[np.isnan(coef)]
       keep = idx[np.isnan(coef)==False]

       # Now actually set dropped ones to zero
       coef[drop] = 0

       # Convert R so that rest of code can just continue as with Chol (i.e., L)
       LP = RP.T.tocsc()

       # Keep only columns of Pr/P that belong to identifiable params. So P.T@LP is Cholesky of
       # negative penalized Hessian of model without unidentifiable coef. Important: LP and Pr/P no
       # longer match dimensions of embedded penalties after this! So we need to keep track of that
       # in the appropriate functions (i.e., `calculate_edf` which calls `compute_B` when called
       # with only LP and not Linv).
       P = P[:,keep]
       _,Pr,_ = translate_sparse(P.tocsc())
       P = compute_eigen_perm(Pr)

    :param y: vector of observations
    :type y: np.ndarray
    :param X: Some rectangular sparse matrix
    :type X: scp.sparse.csc_array
    :param E: Sparse square matrix
    :type E: scp.sparse.csc_array
    :return: Pivoted Cholesky of ``X.T@X + S``, first column pivot indices in a list, second column
        pivot indices in a list, ``b``, estimated rank, and code indicating success.
    :rtype: tuple[scp.sparse.csc_array,list[int],list[int],np.ndarray,int,int]
    """
    return eigen_solvers.solve_coef_pqr(y, *map_csc_to_eigen(X), *map_csc_to_eigen(E))


def cpp_solve_coefXX(
    Xy: np.ndarray, XXS: scp.sparse.csc_array
) -> tuple[scp.sparse.csc_array, list[int], np.ndarray, int]:
    """Solves ``(X.T@X + S)@b = X.T@y`` for ``b`` via sparse Cholesky decomposition with
    ``(X.T@X + S)`` and ``X.T@y`` pre-computed.

    :param Xy: Holds ``X.T@y``
    :type Xy: np.ndarray
    :param XXS: Holds ``(X.T@X + S)``
    :type XXS: scp.sparse.csc_array
    :return: Pivoted Cholesky of ``X.T@X + S``, column pivot indices in a list, ``b``, and
        code indicating success
    :rtype: tuple[scp.sparse.csc_array,list[int],np.ndarray,int]
    """
    return eigen_solvers.solve_coefXX(Xy, *map_csc_to_eigen(XXS))


def cpp_solve_L(
    X: scp.sparse.csc_array, S: scp.sparse.csc_array
) -> tuple[scp.sparse.csc_array, list[int], int]:
    """Solves ``(X.T@X + S)@B=I`` for ``B``, where ``(X.T@X + S)`` is sparse, symmetric, and full
    rank and ``I`` is an identity matrix of suitable dimension via Cholesky decomposition.

    :param X: Some rectangular sparse matrix
    :type X: scp.sparse.csc_array
    :param S: Sparse square matrix
    :type S: scp.sparse.csc_array
    :return: ``B`` (inverse of **pivoted** ``X.T@X + S``), list of pivot indices, and code
        indicating success
    :rtype: tuple[scp.sparse.csc_array,list[int],int]
    """
    return eigen_solvers.solve_L(*map_csc_to_eigen(X), *map_csc_to_eigen(S))


def cpp_solve_LXX(
    A: scp.sparse.csc_array,
) -> tuple[scp.sparse.csc_array, list[int], int]:
    """Solves ``A@B=I`` for ``B``, where ``A`` is sparse, symmetric, and full rank and ``I`` is an
    identity matrix of suitable dimension via Cholesky decomposition.

    :param A: Some sparse symmetric matrix
    :type A: scp.sparse.csc_array
    :return: ``B`` (inverse of **pivoted** ``A``), list of pivot indices, and code indicating
        success
    :rtype: tuple[scp.sparse.csc_array,list[int],int]
    """
    return eigen_solvers.cpp_solve_LXX(*map_csc_to_eigen(A))


def cpp_solve_tr(
    A: scp.sparse.csc_array, C: scp.sparse.csc_array
) -> scp.sparse.csc_array:
    """Solves ``A@B=C``, where ``A`` is sparse and lower triangular. This can be utilized to obtain
    ``B = inv(A)``, when ``C`` is the identity.

    :param A: Lower triangluar sparse matrix
    :type A: scp.sparse.csc_array
    :param C: Sparse potentially rectangular matrix
    :type C: scp.sparse.csc_array
    :return: ``B``
    :rtype: scp.sparse.csc_array
    """
    return eigen_solvers.solve_tr(*map_csc_to_eigen(A), C)


def cpp_backsolve_tr(
    A: scp.sparse.csc_array, C: scp.sparse.csc_array
) -> scp.sparse.csc_array:
    """Solves ``A@B=C``, where ``A`` is sparse and upper triangular. This can be utilized to obtain
    ``B = inv(A)``, when ``C`` is the identity.

    :param A: Lower triangluar sparse matrix
    :type A: scp.sparse.csc_array
    :param C: Sparse potentially rectangular matrix
    :type C: scp.sparse.csc_array
    :return: ``B``
    :rtype: scp.sparse.csc_array
    """
    return eigen_solvers.backsolve_tr(*map_csc_to_eigen(A), C)


def est_condition(
    L: scp.sparse.csc_array,
    Linv: scp.sparse.csc_array,
    seed: int | None = 0,
    verbose: bool = True,
) -> tuple[float, float, float, int]:
    """Estimate the condition number ``K`` - the ratio of the largest to smallest singular values -
    of matrix ``A``, where ``A.T@A = L@L.T``.

    ``L`` and ``Linv`` can either be obtained by Cholesky decomposition, i.e., ``A.T@A = L@L.T`` or
    by QR decomposition ``A=Q@R`` where ``R=L.T``.

    If ``verbose=True`` (default), separate warnings will be issued in case
    ``K>(1/(0.5*sqrt(epsilon)))`` and ``K>(1/(0.5*epsilon))``. If the former warning is raised,
    this indicates that computing ``L`` via a Cholesky decomposition is likely unstable
    and should be avoided. If the second warning is raised as well, obtaining ``L`` via QR
    decomposition (of ``A``) is also likely to be unstable (see Golub & Van Loan, 2013).

    References:
      - Cline et al. (1979). An Estimate for the Condition Number of a Matrix.
      - Golub & Van Loan (2013). Matrix computations, 4th edition.

    :param L: Cholesky or any other root of ``A.T@A`` as a sparse matrix.
    :type L: scp.sparse.csc_array
    :param Linv: Inverse of Choleksy (or any other root) of ``A.T@A``.
    :type Linv: scp.sparse.csc_array
    :param seed: The seed to use for the random parts of the singular value decomposition.
        Defaults to 0.
    :type seed: int or None or numpy.random.Generator
    :param verbose: Whether or not warnings should be printed. Defaults to True.
    :type verbose: bool
    :return: A tuple, containing the estimate of condition number ``K``, an estimate of the largest
        singular value of ``A``, an estimate of the smallest singular value of ``A``, and a
        ``code``. The latter will be zero in case no warning was raised, 1 in case the first
        warning described above was raised, and 2 if the second warning was raised as well.
    :rtype: tuple[float,float,float,int]
    """

    # Get unit round-off (Golub & Van Loan, 2013)
    u = 0.5 * np.finfo(float).eps

    # Now get estimates of largest and smallest singular values of A
    # from norms of L and Linv (Cline et al. 1979)
    try:
        min_sing = scp.sparse.linalg.svds(
            Linv, k=1, return_singular_vectors=False, random_state=seed
        )[0]
        max_sing = scp.sparse.linalg.svds(
            L, k=1, return_singular_vectors=False, random_state=seed
        )[0]
    except:  # noqa: E722
        try:
            min_sing = scp.sparse.linalg.svds(
                Linv,
                k=1,
                return_singular_vectors=False,
                random_state=seed,
                solver="lobpcg",
            )[0]
            max_sing = scp.sparse.linalg.svds(
                L,
                k=1,
                return_singular_vectors=False,
                random_state=seed,
                solver="lobpcg",
            )[0]
        except:  # noqa: E722
            # Solver failed.. get out
            warnings.warn(
                (
                    "Estimating the condition number of matrix A, where A.T@A=L.T@L failed. This "
                    "can happen but might indicate that something is wrong. Consider estimates "
                    "carefully!"
                )
            )
            return np.inf, np.inf, -np.inf, 1

    K = max_sing * min_sing
    code = 0

    if K > 1 / np.sqrt(u):
        if verbose:
            warnings.warn(
                (
                    "Condition number of matrix A, where A.T@A=L.T@L, is larger than 1/sqrt(u), "
                    "where u is half the machine precision."
                )
            )
        code = 1

    if K > 1 / u:
        if verbose:
            warnings.warn(
                (
                    "Condition number of matrix A, where A.T@A=L.T@L, is larger than 1/u, where u "
                    "is half the machine precision."
                )
            )
        code = 2

    return K, max_sing, 1 / min_sing, code


def compute_block_B_shared(
    address_dat: str,
    address_ptr: str,
    address_idx: str,
    shape_dat: tuple,
    shape_ptr: tuple,
    rows: int,
    cols: int,
    nnz: int,
    T: scp.sparse.csc_array,
) -> float:
    """Solves ``L @ B = T`` for ``B`` via forward solving and based on shared memory for ``L``,
    then computes and returns ``B.power(2).sum()``.

    :param address_dat: Address to data array of ``L``
    :type address_dat: str
    :param address_ptr: Address to pointer array of ``L``
    :type address_ptr: str
    :param address_idx: Address to indices array of ``L``
    :type address_idx: str
    :param shape_dat: Shape of data array of ``L``
    :type shape_dat: tuple
    :param shape_ptr: Shape of pointer array of ``L``
    :type shape_ptr: tuple
    :param rows: Number of rows of ``L``
    :type rows: int
    :param cols: Number of cols of ``L``
    :type cols: int
    :param nnz: Number of non-zero elements in ``L``
    :type nnz: int
    :param T: Target matrix
    :type T: scp.sparse.csc_array
    :return: ``B.power(2).sum()``
    :rtype: float
    """
    BB = compute_block_linv_shared(
        address_dat, address_ptr, address_idx, shape_dat, shape_ptr, rows, cols, nnz, T
    )
    return BB.power(2).sum()


def compute_block_B_shared_cluster(
    address_dat: str,
    address_ptr: str,
    address_idx: str,
    shape_dat: tuple,
    shape_ptr: tuple,
    rows: int,
    cols: int,
    nnz: int,
    T: scp.sparse.csc_array,
    cluster_weights: list[float],
) -> tuple[float, float]:
    """Solves ``L @ B = T`` for ``B`` via forward solving and based on shared memory for ``L``,
    then computes and returns ``sum(B.power(2).sum()*cluster_weights)`` and
    ``B.power(2).sum()*len(cluster_weights)``.

    :param address_dat: Address to data array of ``L``
    :type address_dat: str
    :param address_ptr: Address to pointer array of ``L``
    :type address_ptr: str
    :param address_idx: Address to indices array of ``L``
    :type address_idx: str
    :param shape_dat: Shape of data array of ``L``
    :type shape_dat: tuple
    :param shape_ptr: Shape of pointer array of ``L``
    :type shape_ptr: tuple
    :param rows: Number of rows of ``L``
    :type rows: int
    :param cols: Number of cols of ``L``
    :type cols: int
    :param nnz: Number of non-zero elements in ``L``
    :type nnz: int
    :param T: Target matrix
    :type T: scp.sparse.csc_array
    :param cluster_weights: Cluster weights obtained from
        :func:`mssm.src.python.formula.__cluster_discretize`.
    :type cluster_weights: list[float]
    :return: ``sum(B.power(2).sum()*cluster_weights)`` and
        ``B.power(2).sum()*len(cluster_weights)``
    :rtype: tuple[float,float]
    """
    BB = compute_block_linv_shared(
        address_dat, address_ptr, address_idx, shape_dat, shape_ptr, rows, cols, nnz, T
    )
    BBps = BB.power(2).sum()
    return np.sum(cluster_weights * BBps), len(cluster_weights) * BBps


def compute_B(
    L: scp.sparse.csc_array,
    P: scp.sparse.csc_array,
    lTerm: LambdaTerm,
    n_c: int = 10,
    drop: np.typing.NDArray[np.int_] | None = None,
) -> float | tuple[float, float]:
    """Solves ``L @ B = P @ lTerm.D_J_emb`` for ``B``, then returns ``B.power(2).sum()`` or two
    approximations of this (for very big factor smooth models).

    :param L: Lower triangular sparse matrix
    :type L: scp.sparse.csc_array
    :param P: Permuation matrix
    :type P: scp.sparse.csc_array
    :param lTerm: Current penalty term
    :type lTerm: LambdaTerm
    :param n_c: Number of cores, defaults to 10
    :type n_c: int, optional
    :param drop: Array of parameters (columns/rows of ``lTerm.D_J_emb``) to drop, defaults to None
    :type drop: np.typing.NDArray[np.int_] | None, optional
    :return: ``sum(B.power(2).sum()`` or  ``sum(B.power(2).sum()*cluster_weights)`` and
        ``B.power(2).sum()*len(cluster_weights)`` with cluster weights obtained from
        :func:`mssm.src.python.formula.__cluster_discretize`.
    :rtype: float | tuple[float, float]
    """
    # Solves L @ B = P @ D for B, parallelizing over column
    # blocks of D if int(D.shape[1]/2000) > 1

    # Also allows for approximate B computation for very big factor smooths.
    D_start = lTerm.start_index
    idx = np.arange(lTerm.S_J_emb.shape[1])
    if drop is None:
        drop = np.array([])
    keep = idx[np.isin(idx, drop) == False]  # noqa: E712

    if lTerm.clust_series is None:

        col_sums = lTerm.S_J.sum(axis=0)
        if lTerm.type == PenType.NULL and sum(col_sums[col_sums > 0]) == 1:
            # Null penalty for factor smooth has usually only non-zero element in a single colum,
            # so we only need to solve one linear system per level of the factor smooth.
            NULL_idx = np.argmax(col_sums)

            D_idx = np.arange(
                lTerm.start_index + NULL_idx,
                lTerm.S_J.shape[1] * (lTerm.rep_sj + 1),
                lTerm.S_J.shape[1],
            )

            D_len = len(D_idx)
            PD = P @ lTerm.D_J_emb[:, D_idx][keep, :]
        else:
            # First get columns associated to penalty
            D_len = lTerm.rep_sj * lTerm.S_J.shape[1]
            D_end = lTerm.start_index + D_len
            D_idx = idx[D_start:D_end]

            # Now check if dropped column is included, if so remove and update length.
            D_idx = D_idx[np.isin(D_idx, drop) == False]  # noqa: E712
            D_len = len(D_idx)
            PD = P @ lTerm.D_J_emb[:, D_idx][keep, :]

        D_r = int(D_len / 2000)

        if D_r > 1 and n_c > 1:
            # Parallelize over column blocks of P @ D
            # Can speed up computations considerably and is feasible memory-wise
            # since L itself is super sparse.
            n_c = min(D_r, n_c)
            split = np.array_split(range(D_len), n_c)
            PD = P @ lTerm.D_J_emb[:, D_idx][keep, :]
            PDs = [PD[:, split[i]] for i in range(n_c)]

            with (
                managers.SharedMemoryManager() as manager,
                mp.Pool(processes=n_c) as pool,
            ):
                # Create shared memory copies of data, indptr, and indices
                rows, cols, nnz, data, indptr, indices = map_csc_to_eigen(L)
                shape_dat = data.shape
                shape_ptr = indptr.shape

                dat_mem = manager.SharedMemory(data.nbytes)
                dat_shared = np.ndarray(shape_dat, dtype=np.double, buffer=dat_mem.buf)
                dat_shared[:] = data[:]

                ptr_mem = manager.SharedMemory(indptr.nbytes)
                ptr_shared = np.ndarray(shape_ptr, dtype=np.int64, buffer=ptr_mem.buf)
                ptr_shared[:] = indptr[:]

                idx_mem = manager.SharedMemory(indices.nbytes)
                idx_shared = np.ndarray(shape_dat, dtype=np.int64, buffer=idx_mem.buf)
                idx_shared[:] = indices[:]

                args = zip(
                    repeat(dat_mem.name),
                    repeat(ptr_mem.name),
                    repeat(idx_mem.name),
                    repeat(shape_dat),
                    repeat(shape_ptr),
                    repeat(rows),
                    repeat(cols),
                    repeat(nnz),
                    PDs,
                )

                pow_sums = pool.starmap(compute_block_B_shared, args)

            return sum(pow_sums)

        # Not worth parallelizing, solve directly
        B = cpp_solve_tr(L, PD)
        return B.power(2).sum()

    # Approximate the derivative based just on the columns in D_J that belong to the
    # maximum series identified for each cluster. Use the size of the cluster and the weights to
    # correct for the fact that all series in the cluster are slightly different after all.
    if len(drop) > 0:
        raise ValueError(
            "Approximate derivative computation cannot currently handle unidentifiable terms."
        )

    n_coef = lTerm.S_J.shape[0]
    rank = int(lTerm.rank / lTerm.rep_sj)

    sum_bs_lw = 0
    sum_bs_up = 0

    targets = [
        P
        @ lTerm.D_J_emb[
            :,
            (D_start + (s * n_coef)) : (  # noqa: E203
                D_start + ((s + 1) * n_coef) - (n_coef - rank)
            ),
        ]
        for s in lTerm.clust_series
    ]

    if len(targets) < 20 * n_c:

        for weights, target in zip(lTerm.clust_weights, targets):

            BB = cpp_solve_tr(L, target)
            BBps = BB.power(2).sum()
            sum_bs_lw += np.sum(weights * BBps)
            sum_bs_up += len(weights) * BBps

    else:
        # Parallelize
        with managers.SharedMemoryManager() as manager, mp.Pool(processes=n_c) as pool:
            # Create shared memory copies of data, indptr, and indices
            rows, cols, nnz, data, indptr, indices = map_csc_to_eigen(L)
            shape_dat = data.shape
            shape_ptr = indptr.shape

            dat_mem = manager.SharedMemory(data.nbytes)
            dat_shared = np.ndarray(shape_dat, dtype=np.double, buffer=dat_mem.buf)
            dat_shared[:] = data[:]

            ptr_mem = manager.SharedMemory(indptr.nbytes)
            ptr_shared = np.ndarray(shape_ptr, dtype=np.int64, buffer=ptr_mem.buf)
            ptr_shared[:] = indptr[:]

            idx_mem = manager.SharedMemory(indices.nbytes)
            idx_shared = np.ndarray(shape_dat, dtype=np.int64, buffer=idx_mem.buf)
            idx_shared[:] = indices[:]

            args = zip(
                repeat(dat_mem.name),
                repeat(ptr_mem.name),
                repeat(idx_mem.name),
                repeat(shape_dat),
                repeat(shape_ptr),
                repeat(rows),
                repeat(cols),
                repeat(nnz),
                targets,
                lTerm.clust_weights,
            )

            sum_bs_lw_all, sum_bs_up_all = zip(
                *pool.starmap(compute_block_B_shared_cluster, args)
            )

            sum_bs_lw = np.sum(sum_bs_lw_all)
            sum_bs_up = np.sum(sum_bs_up_all)

    return sum_bs_lw, sum_bs_up


def compute_block_linv_shared(
    address_dat: str,
    address_ptr: str,
    address_idx: str,
    shape_dat: tuple,
    shape_ptr: tuple,
    rows: int,
    cols: int,
    nnz: int,
    T: scp.sparse.csc_array,
) -> scp.sparse.csc_array:
    """Solves ``L@B = T`` where ``L`` is available in shared memory and ``T`` is a column subset of
    the identity matrix.

    :param address_dat: Address to data array of ``L``
    :type address_dat: str
    :param address_ptr: Address to pointer array of ``L``
    :type address_ptr: str
    :param address_idx: Address to indices array of ``L``
    :type address_idx: str
    :param shape_dat: Shape of data array of ``L``
    :type shape_dat: tuple
    :param shape_ptr: Shape of pointer array of ``L``
    :type shape_ptr: tuple
    :param rows: Number of rows of ``L``
    :type rows: int
    :param cols: Number of cols of ``L``
    :type cols: int
    :param nnz: Number of non-zero elements in ``L``
    :type nnz: int
    :param T: Target matrix
    :type T: scp.sparse.csc_array
    :return: ``B``
    :rtype: scp.sparse.csc_array
    """
    dat_shared = shared_memory.SharedMemory(name=address_dat, create=False)
    ptr_shared = shared_memory.SharedMemory(name=address_ptr, create=False)
    idx_shared = shared_memory.SharedMemory(name=address_idx, create=False)

    data = np.ndarray(shape_dat, dtype=np.double, buffer=dat_shared.buf)
    indptr = np.ndarray(shape_ptr, dtype=np.int64, buffer=ptr_shared.buf)
    indices = np.ndarray(shape_dat, dtype=np.int64, buffer=idx_shared.buf)

    L = eigen_solvers.solve_tr(rows, cols, nnz, data, indptr, indices, T)

    return L


def compute_Linv(L: scp.sparse.csc_array, n_c: int = 10) -> scp.sparse.csc_array:
    """Solves ``L @ inv(L) = I`` for ``inv(L)`` optionally parallelizing over column blocks of
    ``I``.

    :param L: Lower triangular sparse matrix
    :type L: scp.sparse.csc_array
    :param n_c: Number of cores to use, defaults to 10
    :type n_c: int, optional
    :return: ``inv(L)``
    :rtype: scp.sparse.csc_array
    """
    # Solves L @ inv(L) = I for inv(L) parallelizing over column
    # blocks of I if int(I.shape[1]/2000) > 1

    n_col = L.shape[1]
    r = int(n_col / 2000)
    T = scp.sparse.eye(n_col, format="csc")
    if r > 1 and n_c > 1:
        # Parallelize over column blocks of I
        # Can speed up computations considerably and is feasible memory-wise
        # since L itself is super sparse.

        n_c = min(r, n_c)
        split = np.array_split(range(n_col), n_c)
        Ts = [T[:, split[i]] for i in range(n_c)]

        with managers.SharedMemoryManager() as manager, mp.Pool(processes=n_c) as pool:
            # Create shared memory copies of data, indptr, and indices
            rows, cols, nnz, data, indptr, indices = map_csc_to_eigen(L)
            shape_dat = data.shape
            shape_ptr = indptr.shape

            dat_mem = manager.SharedMemory(data.nbytes)
            dat_shared = np.ndarray(shape_dat, dtype=np.double, buffer=dat_mem.buf)
            dat_shared[:] = data[:]

            ptr_mem = manager.SharedMemory(indptr.nbytes)
            ptr_shared = np.ndarray(shape_ptr, dtype=np.int64, buffer=ptr_mem.buf)
            ptr_shared[:] = indptr[:]

            idx_mem = manager.SharedMemory(indices.nbytes)
            idx_shared = np.ndarray(shape_dat, dtype=np.int64, buffer=idx_mem.buf)
            idx_shared[:] = indices[:]

            args = zip(
                repeat(dat_mem.name),
                repeat(ptr_mem.name),
                repeat(idx_mem.name),
                repeat(shape_dat),
                repeat(shape_ptr),
                repeat(rows),
                repeat(cols),
                repeat(nnz),
                Ts,
            )

            LBinvs = pool.starmap(compute_block_linv_shared, args)

        return scp.sparse.hstack(LBinvs)

    return cpp_solve_tr(L, T)



# flake8: noqa
import mssm
from mssm.models import *
from mssm.src.python.utils import correct_VB, estimateVp
import numpy as np
import os
import copy
import io
from contextlib import redirect_stdout
from mssmViz.sim import *
from .defaults import (
    default_gsmm_test_kwargs,
    max_atol,
    max_rtol,
    init_penalties_tests_gammlss,
    init_penalties_tests_gsmm,
)
from mssm.src.python.mcmc import sample_mssm
from mssm.src.python.formula import build_model_matrix

mssm.src.python.exp_fam.GAUMLSS.init_lambda = init_penalties_tests_gammlss
mssm.src.python.exp_fam.GAMMALS.init_lambda = init_penalties_tests_gammlss
mssm.src.python.exp_fam.MULNOMLSS.init_lambda = init_penalties_tests_gammlss
mssm.src.python.exp_fam.PropHaz.init_lambda = init_penalties_tests_gsmm
mssm.src.python.utils.GAMLSSGSMMFamily.init_lambda = init_penalties_tests_gsmm


class Test_MultivariateHSMM_hard:

    # Start by simulating some data
    np_gen = np.random.default_rng(0)
    n_series = 3
    n_obs = 250
    starts_with_first = False

    # Define True model
    pi = np.array([0.6, 0.2, 0.2])

    T = np.array([[0.0, 0.6, 0.4], [0.3, 0.0, 0.7], [0.6, 0.4, 0.0]])

    mus = [np.array([0.0, 2.5, 5.0]), np.array([-2, 0, -5]), np.array([0, 0, 0])]

    sigmas = [
        np.array([[1, -0.3, 0.5], [-0.3, 1, 0.0], [0.5, 0.0, 1]]),
        np.array([[1, 0.3, -0.5], [0.3, 1, 0.3], [-0.5, 0.3, 1]]),
        np.array([[1, 0.0, 0.2], [0.0, 1, -0.6], [0.2, -0.6, 1]]),
    ]

    # Duration distribution parameters
    mus2 = [17, 12, 7]
    scales = [0.05, 0.05, 0.05]
    mus2_init = [2, 4, 6]
    scales_init = scales

    y_all = [[], [], []]
    states_all = []
    time_all = []
    series_all = []
    durs_all = []
    seed = 0
    for s in range(n_series):

        state = np_gen.choice(np.arange(3), p=pi)

        if starts_with_first:
            alpha = 1 / scales[state]
            beta = alpha / mus2[state]
        else:
            alpha = 1 / scales_init[state]
            beta = alpha / mus2_init[state]

        # duration samples
        dur = int(
            scp.stats.gamma.rvs(a=alpha, scale=(1 / beta), size=1, random_state=seed)[0]
        )
        seed += 1

        y = [[], [], []]
        states = []
        durs = []
        t = 0

        while len(y[0]) < n_obs:
            Y = scp.stats.multivariate_normal.rvs(
                size=dur, mean=mus[state], cov=sigmas[state], random_state=seed
            )
            Y = Y.reshape(dur, 3)
            for m in range(3):
                y[m].extend(Y[:, m])

            for _ in range(dur):
                states.append(state)
            durs.append([state, dur])

            prev_state = state
            state = np_gen.choice(np.arange(3), p=T[state, :])

            alpha = 1 / scales[state]
            beta = alpha / mus2[state]

            # duration samples
            dur = int(
                scp.stats.gamma.rvs(
                    a=alpha, scale=(1 / beta), size=1, random_state=seed
                )[0]
            )
            t += 1
            seed += 1

        for m in range(3):
            y[m] = y[m][:n_obs]
        states = states[:n_obs]
        time = np.arange(n_obs)
        for m in range(3):
            y_all[m].extend(y[m])
        time_all.extend(time)

        for _ in range(n_obs):
            series_all.append(s)

        states_all.append(states)
        durs_all.append(durs)

    dat = pd.DataFrame(
        {
            "y0": y_all[0],
            "y1": y_all[1],
            "y2": y_all[2],
            "time": time_all,
            "series": series_all,
        }
    )

    # Create pseudo-y column for time-varying HsMM
    dat["pseudo"] = [1 for _ in range(dat.shape[0])]

    # Prep dat for hsmm model
    uval, id = np.unique(np.array(dat["series"]), return_index=True)
    sid = np.sort(id)
    tid = np.arange(len(id))

    # Now define models
    o_family = MultiGauss(3, [Identity() for _ in range(3)])
    d_family = GAMMALS(links=[LOG(), LOGb(-0.0001)])
    n_S = 3
    M = 3
    D = 100
    starts_with_first = True

    # Initial parameters
    Y = np.concatenate(
        [
            dat["y0"].values.reshape(-1, 1),
            dat["y1"].values.reshape(-1, 1),
            dat["y2"].values.reshape(-1, 1),
        ],
        axis=1,
    )
    init_size = 50
    init_choice = np_gen.choice(len(dat), size=init_size * n_S, replace=False)
    int_samples = []
    init_mus = []
    init_Rs = []

    for j in range(n_S):
        int_samples.append(Y[init_choice[j * init_size : (j + 1) * init_size]])
        init_mus.append(int_samples[j].mean(axis=0))

        cov = np.cov(int_samples[j].T)
        prec = np.linalg.inv(cov)
        R = np.linalg.cholesky(prec).T
        for m in range(M):
            R[m, m] = np.log(R[m, m])

        init_Rs.extend(R[np.triu_indices(M)])

    # Set up formulas
    ys = []
    Xs = []
    form_n_coef = []
    init_coef = []
    build_mat_idx = []
    build_matrix = []

    obs_families = []
    obs_formulas = []
    links = []
    pars = 0
    extra_coef = 0
    for j in range(n_S):
        obs_families.append([])

        obs_families[-1].append(o_family)
        extra_coef += o_family.extra_coef

        for m in range(3):
            # Model of mean of obs model in state j and of signal m
            obs_formulas.append(Formula(lhs(f"y{m}"), [i()], data=dat))

            init_coef.append(init_mus[j][m])
            form_n_coef.append(obs_formulas[-1].n_coef)
            _ = build_penalties(obs_formulas[-1])

            if j == 0 and m == 0:
                build_matrix.append(True)
                Xs.append(build_model_matrix(obs_formulas[-1]))
            else:
                build_matrix.append(False)
                Xs.append(None)

            build_mat_idx.append(0)
            ys.append(dat[f"y{m}"].values.reshape(-1, 1))

        links.extend(o_family.links)
        pars += 3

    d_dat = pd.DataFrame(
        {"x": [1 for _ in range(len(tid))], "pseudo": [1 for _ in range(len(tid))]}
    )

    d_families = []
    d_formulas = []

    # For time-varying model we need to split here
    d_formulas_tv = []
    tv_Xs = copy.deepcopy(Xs)
    tv_ys = copy.deepcopy(ys)

    for j in range(n_S * (1 if starts_with_first else 2)):
        d_families.append(d_family)

        # Model of mean of dur model in state j
        d_formulas.append(Formula(lhs("pseudo"), [i()], data=d_dat))
        d_formulas_tv.append(Formula(lhs("pseudo"), [i()], data=dat))

        init_coef.append(2.5)
        form_n_coef.append(d_formulas[-1].n_coef)
        _ = build_penalties(d_formulas[-1])
        _ = build_penalties(d_formulas_tv[-1])
        Xs.append(build_model_matrix(d_formulas[-1]))
        ys.append(d_dat["pseudo"].values.reshape(-1, 1))
        tv_Xs.append(build_model_matrix(d_formulas_tv[-1]))
        tv_ys.append(dat["pseudo"].values.reshape(-1, 1))
        build_matrix.append(True)
        build_mat_idx.append(len(build_mat_idx))

        # Model of scale parameter of dur model in state j
        d_formulas.append(Formula(lhs("pseudo"), [i()], data=d_dat))
        d_formulas_tv.append(Formula(lhs("pseudo"), [i()], data=dat))

        init_coef.append(-2)
        form_n_coef.append(d_formulas[-1].n_coef)
        _ = build_penalties(d_formulas[-1])
        _ = build_penalties(d_formulas_tv[-1])
        Xs.append(build_model_matrix(d_formulas[-1]))
        ys.append(d_dat["pseudo"].values.reshape(-1, 1))
        tv_Xs.append(build_model_matrix(d_formulas_tv[-1]))
        tv_ys.append(dat["pseudo"].values.reshape(-1, 1))
        build_matrix.append(True)
        build_mat_idx.append(len(build_mat_idx))

        links.extend(d_family.links)
        pars += 2

    # For hsmm with fixed state transitions and pi we have everything now,
    # so collect copies:
    od_init_coef = copy.deepcopy(init_coef)
    od_form_n_coef = copy.deepcopy(form_n_coef)
    od_build_matrix = copy.deepcopy(build_matrix)
    od_build_mat_idx = copy.deepcopy(build_mat_idx)
    od_pars = pars
    od_links = copy.deepcopy(links)
    od_ys = copy.deepcopy(ys)
    od_Xs = copy.deepcopy(Xs)

    t_formulas = []
    t_formulas_tv = []
    for j in range(n_S):
        for par in range(n_S - 2):

            # Model of state transition away from state j
            t_formulas.append(Formula(lhs("pseudo"), [i()], data=d_dat))
            t_formulas_tv.append(Formula(lhs("pseudo"), [i()], data=dat))

            init_coef.append(np.log(1))
            form_n_coef.append(t_formulas[-1].n_coef)
            _ = build_penalties(t_formulas[-1])
            _ = build_penalties(t_formulas_tv[-1])
            Xs.append(build_model_matrix(t_formulas[-1]))
            ys.append(d_dat["pseudo"].values.reshape(-1, 1))
            tv_Xs.append(build_model_matrix(t_formulas_tv[-1]))
            tv_ys.append(dat["pseudo"].values.reshape(-1, 1))
            build_matrix.append(True)
            build_mat_idx.append(len(build_mat_idx))
            pars += 1

    pi_formulas = []
    pi_formulas_tv = []
    for par in range(n_S - 1):
        # Model of initial state probability
        pi_formulas.append(Formula(lhs("pseudo"), [i()], data=d_dat))
        pi_formulas_tv.append(Formula(lhs("pseudo"), [i()], data=dat))

        init_coef.append(np.log(1))
        form_n_coef.append(pi_formulas[-1].n_coef)
        _ = build_penalties(pi_formulas[-1])
        _ = build_penalties(pi_formulas_tv[-1])
        Xs.append(build_model_matrix(pi_formulas[-1]))
        ys.append(d_dat["pseudo"].values.reshape(-1, 1))
        tv_Xs.append(build_model_matrix(pi_formulas_tv[-1]))
        tv_ys.append(dat["pseudo"].values.reshape(-1, 1))
        build_matrix.append(True)
        build_mat_idx.append(len(build_mat_idx))
        pars += 1

    # Initialize Families
    fam1 = HSMMFamily(
        pars,
        links,
        n_S,
        obs_fams=obs_families,
        d_fams=d_families,
        sid=sid,
        tid=tid,
        D=D,
        M=M,
        starts_with_first=starts_with_first,
        ends_with_last=False,
        ends_in_last=False,
        n_cores=1,
        build_mat_idx=build_mat_idx,
    )

    # Time-varying version
    fam2 = HSMMFamily(
        pars,
        links,
        n_S,
        obs_fams=obs_families,
        d_fams=d_families,
        sid=sid,
        tid=None,  # Can set to None since it's same as sid
        D=D,
        M=M,
        starts_with_first=starts_with_first,
        ends_with_last=False,
        ends_in_last=False,
        n_cores=1,
        build_mat_idx=build_mat_idx,
    )

    # Fixed pi and T
    fam3 = HSMMFamily(
        od_pars,
        od_links,
        n_S,
        obs_fams=obs_families,
        d_fams=d_families,
        sid=sid,
        tid=tid,
        D=D,
        M=M,
        starts_with_first=starts_with_first,
        ends_with_last=False,
        ends_in_last=False,
        T=T,
        pi=pi,
        n_cores=1,
        build_mat_idx=od_build_mat_idx,
    )

    # Initialize penalties and coef

    # Extra coef for covariances
    od_form_n_coef.append(extra_coef)
    form_n_coef.append(extra_coef)
    od_init_coef.extend(init_Rs)
    init_coef.extend(init_Rs)

    coef_split_idx = form_n_coef[:-1]

    for coef_i in range(1, len(coef_split_idx)):
        coef_split_idx[coef_i] += coef_split_idx[coef_i - 1]

    od_coef_split_idx = od_form_n_coef[:-1]

    for coef_i in range(1, len(od_coef_split_idx)):
        od_coef_split_idx[coef_i] += od_coef_split_idx[coef_i - 1]

    total_coef = np.sum(form_n_coef)
    od_total_coef = np.sum(od_form_n_coef)
    init_coef = np.array(init_coef).reshape(-1, 1)
    od_init_coef = np.array(od_init_coef).reshape(-1, 1)

    def test_llk(self):
        # Test llk
        llk1 = self.fam1.llk(self.init_coef, self.coef_split_idx, self.ys, self.Xs)
        llk2 = self.fam2.llk(
            self.init_coef, self.coef_split_idx, self.tv_ys, self.tv_Xs
        )
        llk3 = self.fam3.llk(
            self.od_init_coef, self.od_coef_split_idx, self.od_ys, self.od_Xs
        )

        np.testing.assert_allclose(
            llk1,
            llk2,
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

        np.testing.assert_allclose(
            llk1,
            -2422.1207972824186,
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

        np.testing.assert_allclose(
            llk3,
            -2422.408060704047,
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

    def test_grad(self):
        # Test grad function of hsmm and time-varying hsmm
        grad1 = self.fam1.gradient(
            self.init_coef, self.coef_split_idx, self.ys, self.Xs
        )
        grad2 = self.fam2.gradient(
            self.init_coef, self.coef_split_idx, self.tv_ys, self.tv_Xs
        )

        pos_llk_warp = lambda x: self.fam1.llk(
            x.reshape(-1, 1), self.coef_split_idx, self.ys, self.Xs
        )

        grad3 = scp.optimize.approx_fprime(self.init_coef.flatten(), pos_llk_warp)

        grad_diff1 = np.abs(np.round(grad1 - grad3.reshape(-1, 1), decimals=3)).max()
        grad_diff2 = np.abs(np.round(grad2 - grad3.reshape(-1, 1), decimals=3)).max()

        np.testing.assert_allclose(
            grad_diff1,
            0.0,
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

        np.testing.assert_allclose(
            grad_diff2,
            0.0,
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

    def test_grad_fixed(self):
        # Test grad function of fixed hsmm
        grad4 = self.fam3.gradient(
            self.od_init_coef, self.od_coef_split_idx, self.od_ys, self.od_Xs
        )
        pos_llk_warp = lambda x: self.fam3.llk(
            x.reshape(-1, 1), self.od_coef_split_idx, self.od_ys, self.od_Xs
        )
        grad5 = scp.optimize.approx_fprime(self.od_init_coef.flatten(), pos_llk_warp)

        grad_diff3 = np.abs(np.round(grad4 - grad5.reshape(-1, 1), decimals=3)).max()

        np.testing.assert_allclose(
            grad_diff3,
            0.0,
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

    def test_ods(self):
        # Check observation and duration probs
        ods1 = self.fam1.compute_od_probs(
            self.init_coef, self.coef_split_idx, self.ys, self.Xs
        )

        ods2 = self.fam2.compute_od_probs(
            self.init_coef, self.coef_split_idx, self.tv_ys, self.tv_Xs
        )

        ods3 = self.fam3.compute_od_probs(
            self.od_init_coef, self.od_coef_split_idx, self.od_ys, self.od_Xs
        )

        np.testing.assert_allclose(
            ods1[0][0],
            ods2[0][0],
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

        np.testing.assert_allclose(
            ods1[0][1],
            ods2[0][1][:, :, 0],
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

        np.testing.assert_allclose(
            ods3[0][0][:5, :],
            np.array(
                [
                    [-4.40202288, -3.8485654, -4.34099828],
                    [-3.8074954, -3.87171952, -3.6856842],
                    [-5.10074097, -4.5286156, -4.96309288],
                    [-3.10149621, -3.24096177, -3.15586644],
                    [-3.06516362, -2.61534322, -2.90384498],
                ]
            ),
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

        np.testing.assert_allclose(
            ods3[0][1][:5, :],
            np.array(
                [
                    [-11.61207055, -11.61207055, -11.61207055],
                    [-7.79337881, -7.79337881, -7.79337881],
                    [-5.81113444, -5.81113444, -5.81113444],
                    [-4.58076989, -4.58076989, -4.58076989],
                    [-3.76239347, -3.76239347, -3.76239347],
                ]
            ),
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

    def test_tps(self):
        # Check Ts and ps
        Tps1 = self.fam1.compute_Tpi(
            self.init_coef, self.coef_split_idx, self.ys, self.Xs
        )

        Tps2 = self.fam2.compute_Tpi(
            self.init_coef, self.coef_split_idx, self.tv_ys, self.tv_Xs
        )

        Tps3 = self.fam3.compute_Tpi(
            self.od_init_coef, self.od_coef_split_idx, self.od_ys, self.od_Xs
        )

        np.testing.assert_allclose(
            Tps1[0][0],
            Tps2[0][0][:, :, 0],
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

        np.testing.assert_allclose(
            Tps1[0][1],
            Tps2[0][1],
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

        np.testing.assert_allclose(
            Tps3[0][0],
            self.T,
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

        np.testing.assert_allclose(
            Tps3[0][1],
            self.pi,
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

    def test_viterbi(self):
        # Test viterbi
        viterbi1 = self.fam1.decode_viterbi(
            self.init_coef, self.coef_split_idx, self.ys, self.Xs
        )

        viterbi2 = self.fam2.decode_viterbi(
            self.init_coef, self.coef_split_idx, self.tv_ys, self.tv_Xs
        )

        viterbi3 = self.fam3.decode_viterbi(
            self.od_init_coef, self.od_coef_split_idx, self.od_ys, self.od_Xs
        )

        np.testing.assert_allclose(
            viterbi1[0][1],
            viterbi2[0][1],
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

        np.testing.assert_allclose(
            viterbi1[0][1][:20],
            viterbi3[0][1][:20],
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

        np.testing.assert_allclose(
            viterbi1[0][1][:20],
            np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2]),
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

    def test_state_samples(self):
        # test state samples
        state_samples1 = self.fam1.sample_posterior_states(
            self.init_coef, self.coef_split_idx, self.ys, self.Xs, 2
        )

        state_samples2 = self.fam2.sample_posterior_states(
            self.init_coef, self.coef_split_idx, self.tv_ys, self.tv_Xs, 2
        )

        state_samples3 = self.fam3.sample_posterior_states(
            self.od_init_coef, self.od_coef_split_idx, self.od_ys, self.od_Xs, 2
        )

        np.testing.assert_allclose(
            state_samples1[0][1],
            state_samples2[0][1],
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

        np.testing.assert_allclose(
            state_samples1[0][1][:20, 1],
            np.array([1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0]),
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

        np.testing.assert_allclose(
            state_samples3[0][1][:20, 1],
            np.array([0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0]),
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

    def test_predict(self):
        # test prediction
        prediction1 = self.fam1.predict(
            self.init_coef, self.coef_split_idx, self.ys, self.Xs, n_samples=2
        )

        prediction2 = self.fam2.predict(
            self.init_coef, self.coef_split_idx, self.tv_ys, self.tv_Xs, n_samples=2
        )

        prediction3 = self.fam3.predict(
            self.od_init_coef,
            self.od_coef_split_idx,
            self.od_ys,
            self.od_Xs,
            n_samples=2,
        )

        np.testing.assert_allclose(
            prediction1[0][0],
            prediction2[0][0],
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

        np.testing.assert_allclose(
            prediction1[0][1],
            prediction2[0][1],
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

        np.testing.assert_allclose(
            prediction1[0][0][:5, :, 0],
            np.array(
                [
                    [-2.96476414, -0.65006362, -7.60078721],
                    [0.25008563, 3.18114284, 1.81762191],
                    [-2.22843751, -1.68034947, -8.39775076],
                    [0.04533482, 1.47884745, -0.24022577],
                    [-1.90705985, -1.56167648, -1.1013139],
                ]
            ),
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

        np.testing.assert_allclose(
            prediction3[0][0][:5, :, 0],
            np.array(
                [
                    [-2.96476414, -0.65006362, -7.60078721],
                    [0.25008563, 3.18114284, 1.81762191],
                    [-2.22843751, -1.68034947, -8.39775076],
                    [0.04533482, 1.47884745, -0.24022577],
                    [-1.90705985, -1.56167648, -1.1013139],
                ]
            ),
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

    def test_predictive_resid(self):
        # Test predictive resid
        pred_resid1 = self.fam1.get_resid(
            self.init_coef,
            self.coef_split_idx,
            self.ys,
            self.Xs,
            resid_type="predictive",
        )

        pred_resid2 = self.fam2.get_resid(
            self.init_coef,
            self.coef_split_idx,
            self.tv_ys,
            self.tv_Xs,
            resid_type="predictive",
        )

        pred_resid3 = self.fam3.get_resid(
            self.od_init_coef,
            self.od_coef_split_idx,
            self.od_ys,
            self.od_Xs,
            resid_type="predictive",
        )

        np.testing.assert_allclose(
            pred_resid1,
            pred_resid2,
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

        np.testing.assert_allclose(
            pred_resid1[:5, :],
            np.array(
                [
                    [-2.85046777, -1.7181105, -4.19853279],
                    [-0.7696748, -1.30788421, -1.40362903],
                    [-1.25885595, -1.64247998, -1.72167249],
                    [-1.24941138, -1.66806177, -1.99060636],
                    [-1.5177543, -1.92540368, -2.2091171],
                ]
            ),
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

        np.testing.assert_allclose(
            pred_resid3[:5, :],
            np.array(
                [
                    [-2.81573851, -1.71229736, -4.08310188],
                    [-0.75344712, -1.3088929, -1.4026881],
                    [-1.24805185, -1.64312734, -1.72083857],
                    [-1.23201096, -1.67057077, -1.9899709],
                    [-1.50021403, -1.92840169, -2.20746248],
                ]
            ),
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

    def test_forward_resid(self):
        # Test forward resid
        forward_resid1 = self.fam1.get_resid(
            self.init_coef,
            self.coef_split_idx,
            self.ys,
            self.Xs,
            resid_type="forward",
        )

        forward_resid2 = self.fam2.get_resid(
            self.init_coef,
            self.coef_split_idx,
            self.tv_ys,
            self.tv_Xs,
            resid_type="forward",
        )

        forward_resid3 = self.fam3.get_resid(
            self.od_init_coef,
            self.od_coef_split_idx,
            self.od_ys,
            self.od_Xs,
            resid_type="forward",
        )

        np.testing.assert_allclose(
            forward_resid1,
            forward_resid2,
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

        np.testing.assert_allclose(
            forward_resid1[:5, :],
            np.array(
                [
                    [-1.94598191, -0.65367552, -0.96855481],
                    [1.56183752, 0.64444896, -1.20854879],
                    [-2.26434264, -0.52807118, -0.9964072],
                    [0.97091134, 1.09212051, -0.99332044],
                    [-0.26832785, 0.5483176, -1.38903507],
                ]
            ),
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

        np.testing.assert_allclose(
            forward_resid3[:5, :],
            np.array(
                [
                    [-1.97745873, -0.62443529, -0.9983917],
                    [1.53899815, 0.59128775, -1.25965858],
                    [-2.29025925, -0.51067471, -1.03167802],
                    [0.97242687, 1.0164081, -1.02992055],
                    [-0.28974535, 0.50511576, -1.44570645],
                ]
            ),
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

    def test_viterbi_resid(self):
        # Test viterbi resid
        viterbi_resid1 = self.fam1.get_resid(
            self.init_coef,
            self.coef_split_idx,
            self.ys,
            self.Xs,
            resid_type="viterbi_dur",
        )

        viterbi_resid2 = self.fam2.get_resid(
            self.init_coef,
            self.coef_split_idx,
            self.tv_ys,
            self.tv_Xs,
            resid_type="viterbi_dur",
        )

        viterbi_resid3 = self.fam3.get_resid(
            self.od_init_coef,
            self.od_coef_split_idx,
            self.od_ys,
            self.od_Xs,
            resid_type="viterbi_dur",
        )

        np.testing.assert_allclose(
            viterbi_resid1,
            viterbi_resid2,
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

        np.testing.assert_allclose(
            viterbi_resid1,
            np.array(
                [
                    [0.26190762, 0.01786241, 0.523248],
                    [1.01132804, 0.85198392, 0.58350554],
                    [0.50689602, 0.76070477, 0.74922525],
                ]
            ),
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

        np.testing.assert_allclose(
            viterbi_resid3,
            np.array(
                [
                    [0.79597972, 1.0079424, 0.70741119],
                    [0.63609553, 0.73972272, 0.4817017],
                    [0.50689602, 0.76070477, 0.74922525],
                ]
            ),
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

    def test_posterior_resid(self):
        # Test posterior resid
        posterior_resid1 = self.fam1.get_resid(
            self.init_coef,
            self.coef_split_idx,
            self.ys,
            self.Xs,
            resid_type="posterior_dur",
        )

        posterior_resid2 = self.fam2.get_resid(
            self.init_coef,
            self.coef_split_idx,
            self.tv_ys,
            self.tv_Xs,
            resid_type="posterior_dur",
        )

        posterior_resid3 = self.fam3.get_resid(
            self.od_init_coef,
            self.od_coef_split_idx,
            self.od_ys,
            self.od_Xs,
            resid_type="posterior_dur",
        )

        np.testing.assert_allclose(
            posterior_resid1,
            posterior_resid2,
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

        np.testing.assert_allclose(
            posterior_resid1[:, :, 0],
            np.array(
                [
                    [0.16469857, -0.4488697, 0.13306259],
                    [0.19098355, 0.11323988, 0.01077971],
                    [-0.39889136, -0.39197353, -0.53289776],
                ]
            ),
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

        np.testing.assert_allclose(
            posterior_resid3[:, :, 0],
            np.array(
                [
                    [0.39645963, -0.11576908, 0.22357649],
                    [0.2105846, -0.23401751, -0.20967159],
                    [-0.08712459, -0.5551245, -0.41983202],
                ]
            ),
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )


class Test_UnivariateHSMM_hard:

    # Start by simulating some data - same as above
    np_gen = np.random.default_rng(0)
    n_series = 3
    n_obs = 250
    starts_with_first = False

    # Define True model
    pi = np.array([0.6, 0.2, 0.2])

    T = np.array([[0.0, 0.6, 0.4], [0.3, 0.0, 0.7], [0.6, 0.4, 0.0]])

    mus = [np.array([0.0, 2.5, 5.0]), np.array([-2, 0, -5]), np.array([0, 0, 0])]

    sigmas = [
        np.array([[1, -0.3, 0.5], [-0.3, 1, 0.0], [0.5, 0.0, 1]]),
        np.array([[1, 0.3, -0.5], [0.3, 1, 0.3], [-0.5, 0.3, 1]]),
        np.array([[1, 0.0, 0.2], [0.0, 1, -0.6], [0.2, -0.6, 1]]),
    ]

    # Duration distribution parameters
    mus2 = [17, 12, 7]
    scales = [0.05, 0.05, 0.05]
    mus2_init = [2, 4, 6]
    scales_init = scales

    y_all = [[], [], []]
    states_all = []
    time_all = []
    series_all = []
    durs_all = []
    seed = 0
    for s in range(n_series):

        state = np_gen.choice(np.arange(3), p=pi)

        if starts_with_first:
            alpha = 1 / scales[state]
            beta = alpha / mus2[state]
        else:
            alpha = 1 / scales_init[state]
            beta = alpha / mus2_init[state]

        # duration samples
        dur = int(
            scp.stats.gamma.rvs(a=alpha, scale=(1 / beta), size=1, random_state=seed)[0]
        )
        seed += 1

        y = [[], [], []]
        states = []
        durs = []
        t = 0

        while len(y[0]) < n_obs:
            Y = scp.stats.multivariate_normal.rvs(
                size=dur, mean=mus[state], cov=sigmas[state], random_state=seed
            )
            Y = Y.reshape(dur, 3)
            for m in range(3):
                y[m].extend(Y[:, m])

            for _ in range(dur):
                states.append(state)
            durs.append([state, dur])

            prev_state = state
            state = np_gen.choice(np.arange(3), p=T[state, :])

            alpha = 1 / scales[state]
            beta = alpha / mus2[state]

            # duration samples
            dur = int(
                scp.stats.gamma.rvs(
                    a=alpha, scale=(1 / beta), size=1, random_state=seed
                )[0]
            )
            t += 1
            seed += 1

        for m in range(3):
            y[m] = y[m][:n_obs]
        states = states[:n_obs]
        time = np.arange(n_obs)
        for m in range(3):
            y_all[m].extend(y[m])
        time_all.extend(time)

        for _ in range(n_obs):
            series_all.append(s)

        states_all.append(states)
        durs_all.append(durs)

    dat = pd.DataFrame(
        {
            "y0": y_all[0],
            "y1": y_all[1],
            "y2": y_all[2],
            "time": time_all,
            "series": series_all,
        }
    )

    # Prep dat for hsmm model
    uval, id = np.unique(np.array(dat["series"]), return_index=True)
    sid = np.sort(id)
    tid = np.arange(len(id))

    # Now set up model
    o_family = GAUMLSS(links=[Identity(), LOGb(-0.0001)])

    d_family = GAMMALS(links=[LOG(), LOGb(-0.0001)])
    n_S = 3
    M = 3
    D = 100
    starts_with_first = False

    # Initial parameters
    Y = np.concatenate(
        [
            dat["y0"].values.reshape(-1, 1),
            dat["y1"].values.reshape(-1, 1),
            dat["y2"].values.reshape(-1, 1),
        ],
        axis=1,
    )
    init_size = 50
    init_choice = np_gen.choice(len(dat), size=init_size * n_S, replace=False)
    int_samples = []
    init_mus = []
    init_sds = []

    for j in range(n_S):
        int_samples.append(Y[init_choice[j * init_size : (j + 1) * init_size]])
        init_mus.append(int_samples[j].mean(axis=0))

        cov = np.cov(int_samples[j].T)
        init_sds.append(np.diag(cov))

    # Set up formulas
    ys = []
    Xs = []
    form_n_coef = []
    init_coef = []
    build_mat_idx = []
    build_matrix = []

    obs_families = []
    obs_formulas = []
    links = []
    pars = 0
    extra_coef = 0
    for j in range(n_S):
        obs_families.append([])

        for m in range(M):
            obs_families[-1].append(o_family)
            # Model of mean of obs model in state j and of signal m
            obs_formulas.append(Formula(lhs(f"y{m}"), [i()], data=dat))

            init_coef.append(init_mus[j][m])
            form_n_coef.append(obs_formulas[-1].n_coef)
            _ = build_penalties(obs_formulas[-1])
            Xs.append(build_model_matrix(obs_formulas[-1]))
            ys.append(dat[f"y{m}"].values.reshape(-1, 1))

            build_matrix.append(True)
            build_mat_idx.append(len(build_mat_idx))

            # Model of scale parameter of obs model in state j and of signal m
            obs_formulas.append(Formula(lhs(f"y{m}"), [i()], data=dat))

            init_coef.append(np.log(np.sqrt(init_sds[j][m])))
            form_n_coef.append(obs_formulas[-1].n_coef)
            _ = build_penalties(obs_formulas[-1])
            Xs.append(build_model_matrix(obs_formulas[-1]))
            ys.append(dat[f"y{m}"].values.reshape(-1, 1))

            build_matrix.append(True)
            build_mat_idx.append(len(build_mat_idx))

            links.extend(o_family.links)
            pars += 2

    d_dat = pd.DataFrame(
        {"x": [1 for _ in range(len(tid))], "y": [1 for _ in range(len(tid))]}
    )

    d_families = []
    d_formulas = []

    for j in range(n_S * (1 if starts_with_first else 2)):
        d_families.append(d_family)

        # Model of mean of dur model in state j
        d_formulas.append(Formula(lhs("y"), [i()], data=d_dat))

        init_coef.append(2.5)
        form_n_coef.append(d_formulas[-1].n_coef)
        _ = build_penalties(d_formulas[-1])
        Xs.append(build_model_matrix(d_formulas[-1]))
        ys.append(d_dat["y"].values.reshape(-1, 1))
        build_matrix.append(True)
        build_mat_idx.append(len(build_mat_idx))

        # Model of scale parameter of dur model in state j
        d_formulas.append(Formula(lhs("y"), [i()], data=d_dat))

        init_coef.append(-2)
        form_n_coef.append(d_formulas[-1].n_coef)
        _ = build_penalties(d_formulas[-1])
        Xs.append(build_model_matrix(d_formulas[-1]))
        ys.append(d_dat["y"].values.reshape(-1, 1))
        build_matrix.append(True)
        build_mat_idx.append(len(build_mat_idx))

        links.extend(d_family.links)
        pars += 2

    t_formulas = []
    for j in range(n_S):
        for par in range(n_S - 2):

            # Model of state transition away from state j
            t_formulas.append(Formula(lhs("y"), [i()], data=d_dat))

            init_coef.append(np.log(1))
            form_n_coef.append(t_formulas[-1].n_coef)
            _ = build_penalties(t_formulas[-1])
            Xs.append(build_model_matrix(t_formulas[-1]))
            ys.append(d_dat["y"].values.reshape(-1, 1))
            build_matrix.append(True)
            build_mat_idx.append(len(build_mat_idx))
            pars += 1

    pi_formulas = []

    for par in range(n_S - 1):
        # Model of initial state probability
        pi_formulas.append(Formula(lhs("y"), [i()], data=d_dat))

        init_coef.append(np.log(1))
        form_n_coef.append(pi_formulas[-1].n_coef)
        _ = build_penalties(pi_formulas[-1])
        Xs.append(build_model_matrix(pi_formulas[-1]))
        ys.append(d_dat["y"].values.reshape(-1, 1))
        build_matrix.append(True)
        build_mat_idx.append(len(build_mat_idx))
        pars += 1

    # Initialize Family
    fam = HSMMFamily(
        pars,
        links,
        n_S,
        obs_fams=obs_families,
        d_fams=d_families,
        sid=sid,
        tid=tid,
        D=D,
        M=M,
        starts_with_first=starts_with_first,
        ends_with_last=False,
        ends_in_last=False,
        n_cores=1,
        build_mat_idx=build_mat_idx,
    )

    # Initialize coef and penalties
    coef_split_idx = form_n_coef[:-1]

    for coef_i in range(1, len(coef_split_idx)):
        coef_split_idx[coef_i] += coef_split_idx[coef_i - 1]

    total_coef = np.sum(form_n_coef)
    init_coef = np.array(init_coef).reshape(-1, 1)

    def test_llk(self):
        # Test log-likelihood of univariate hsmm
        llk = self.fam.llk(self.init_coef, self.coef_split_idx, self.ys, self.Xs)

        np.testing.assert_allclose(
            llk,
            -4870.823813514383,
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

    def test_grad(self):
        # Test grad function of univariate hsmm
        grad = self.fam.gradient(self.init_coef, self.coef_split_idx, self.ys, self.Xs)
        pos_llk_warp = lambda x: self.fam.llk(
            x.reshape(-1, 1), self.coef_split_idx, self.ys, self.Xs
        )
        grad2 = scp.optimize.approx_fprime(self.init_coef.flatten(), pos_llk_warp)

        grad_diff = np.abs(np.round(grad - grad2.reshape(-1, 1), decimals=3)).max()

        np.testing.assert_allclose(
            grad_diff,
            0.0,
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )


class Test_UnivariateHMP:
    # Start by simulating some data
    np_gen = np.random.default_rng(0)
    n_series = 100

    # Define True model
    pi = None

    T = None

    mus = [np.array([0.0, 2.5, 5.0]), np.array([-2, 0, -5])]

    # Duration distribution parameters
    mus2 = [17, 12, 7]
    scales = [0.05, 0.05, 0.05]

    y_all = [[], [], []]
    states_all = []
    time_all = []
    series_all = []
    seed = 0
    durs_all = []
    for s in range(n_series):
        y = [[], [], []]
        states = []
        durs = []
        t = 0
        state = 0

        while state < 5:

            if state % 2 == 1:
                dur = 5
                Y = scp.stats.multivariate_normal.rvs(
                    size=dur,
                    mean=mus[state // 2],
                    cov=np.identity(3),
                    random_state=seed,
                )

            else:
                alpha = 1 / scales[state // 2]
                beta = alpha / mus2[state // 2]

                # duration samples
                dur = int(
                    scp.stats.gamma.rvs(
                        a=alpha, scale=(1 / beta), size=1, random_state=seed
                    )[0]
                )

                Y = scp.stats.multivariate_normal.rvs(
                    size=dur, mean=np.zeros(3), cov=np.identity(3), random_state=seed
                )

            seed += 1

            for m in range(3):
                y[m].extend(Y[:, m])

            for _ in range(dur):
                states.append(state)

            state += 1

        n_obs = len(states)
        time = np.arange(n_obs)
        time_all.extend(time)
        durs_all.append(len(time))

        for _ in range(n_obs):
            series_all.append(s)

        for m in range(3):
            y_all[m].extend(y[m])

        states_all.append(states)

    dat = pd.DataFrame(
        {
            "y0": y_all[0],
            "y1": y_all[1],
            "y2": y_all[2],
            "time": time_all,
            "series": series_all,
        }
    )

    dat["x"] = np_gen.random(size=len(time_all))

    # Prep dat for hsmm model
    uval, id = np.unique(np.array(dat["series"]), return_index=True)
    sid = np.sort(id)
    tid = np.arange(len(id))

    # Event shape assumed by Anderson et al. (2016)
    event_shape = np.array([0.30901699, 0.80901699, 1.0, 0.80901699, 0.30901699])

    # Test gradients
    ars = [False, True]
    shared_ms = [False, True]
    init_coefs = []
    coef_split_idxs = []
    yss = []
    Xss = []
    families = []

    for ar in ars:
        for shared_m in shared_ms:

            o_family = None
            d_family = GAMMALS(links=[LOG(), LOGb(-0.0001)])
            n_S = 5
            M = 3
            D = max(durs_all)

            ys = []
            Xs = []
            form_n_coef = []
            init_coef = []

            obs_families = []
            shared_formulas = []
            obs_formulas = []
            build_matrix = []
            build_mat_idx = []
            links = []
            pars = 0
            shared_pars = [0]

            rho = 0.5
            Lrhoi = None

            if ar:
                ar_form = Formula(lhs(f"y{1}"), [i()], data=dat, series_id="series")
                Lrhoi, _ = computeAr1Chol(ar_form, rho)

            if len(shared_pars) > 0:
                midx = 1
                if shared_m is False:
                    midx = M

                for m in range(midx):

                    shared_formulas.append(Formula(lhs(f"y{m}"), [f(["x"])], data=dat))

                    init_coef.extend([0.01 for _ in range(shared_formulas[-1].n_coef)])
                    form_n_coef.append(shared_formulas[-1].n_coef)
                    _ = build_penalties(shared_formulas[-1])

                    ys.append(dat[f"y{m}"].values.reshape(-1, 1))

                    if m == 0:
                        build_matrix.append(True)
                        Xs.append(build_model_matrix(shared_formulas[-1]))
                    else:
                        build_matrix.append(False)
                        Xs.append(None)
                    build_mat_idx.append(0)

            event = 0
            start_sep = len(build_matrix)
            for j in range(n_S):
                obs_families.append([])

                for m in range(M):

                    obs_families[-1].append(o_family)

                    if j % 2 == 1:

                        # Model of mean of obs model in state j and of signal m
                        obs_formulas.append(Formula(lhs(f"y{m}"), [i()], data=dat))

                        init_coef.extend(
                            scp.stats.norm.rvs(
                                size=obs_formulas[-1].n_coef,
                                random_state=event * (m + 1) + m,
                            )
                        )
                        form_n_coef.append(obs_formulas[-1].n_coef)
                        _ = build_penalties(obs_formulas[-1])
                        ys.append(dat[f"y{m}"].values.reshape(-1, 1))

                        if len(build_matrix) == start_sep:
                            build_matrix.append(True)
                            Xs.append(build_model_matrix(obs_formulas[-1]))
                        else:
                            build_matrix.append(False)
                            Xs.append(None)
                        build_mat_idx.append(start_sep)

                        links.extend([Identity()])
                        if m >= (M - 1):
                            event += 1
                        pars += 1

            d_dat = pd.DataFrame({"d": np.arange(len(sid)), "sub": np.ones(len(sid))})

            d_dat = d_dat.astype({"sub": "O"})

            d_families = []
            d_formulas = []

            flat = 0
            start_durs = len(build_mat_idx)
            for j in range(n_S):

                if j % 2 == 1:
                    d_families.append(None)
                else:

                    d_families.append(d_family)

                    # Model of mean of dur model in state j
                    d_formulas.append(Formula(lhs("d"), [ri("sub", id=1)], data=d_dat))

                    build_matrix.append(True)
                    _ = build_penalties(d_formulas[-1])
                    Xs.append(build_model_matrix(d_formulas[-1]))
                    form_n_coef.append(d_formulas[-1].n_coef)
                    init_coef.extend([5])
                    ys.append(d_dat["d"].values.reshape(-1, 1))
                    build_mat_idx.append(len(build_mat_idx))

                    # Model of scale of dur model in state j
                    d_formulas.append(Formula(lhs("d"), [ri("sub", id=2)], data=d_dat))

                    build_matrix.append(True)
                    _ = build_penalties(d_formulas[-1])
                    Xs.append(build_model_matrix(d_formulas[-1]))
                    form_n_coef.append(d_formulas[-1].n_coef)
                    init_coef.extend([1])
                    ys.append(d_dat["d"].values.reshape(-1, 1))
                    build_mat_idx.append(len(build_mat_idx))

                    links.extend(d_family.links)
                    flat += 1
                    pars += 2

            # Initialize Family
            T = np.diag(
                np.ones(n_S - 1), 1
            )  # Super diagonal shift matrix for transitions

            pi = np.zeros(n_S)  # Start in first state
            pi[0] = 1
            fam = HSMMFamily(
                pars,
                links,
                n_S,
                obs_fams=obs_families,
                d_fams=d_families,
                sid=sid,
                tid=tid,
                D=D,
                M=M,
                starts_with_first=True,
                ends_with_last=True,
                ends_in_last=True,
                T=T,
                pi=pi,
                scale=1,
                event_template=event_shape,
                hmp_fam="Gaussian",
                n_cores=4,
                build_mat_idx=build_mat_idx,
                shared_pars=shared_pars,
                shared_m=shared_m,
                Lrhoi=Lrhoi,
            )

            # Setup coef
            coef_split_idx = form_n_coef[:-1]

            for coef_i in range(1, len(coef_split_idx)):
                coef_split_idx[coef_i] += coef_split_idx[coef_i - 1]

            total_coef = np.sum(form_n_coef)
            init_coef = np.array(init_coef).reshape(-1, 1)

            # Collect
            init_coefs.append(copy.deepcopy(init_coef))
            coef_split_idxs.append(copy.deepcopy(coef_split_idx))
            yss.append(copy.deepcopy(ys))
            Xss.append(copy.deepcopy(Xs))
            families.append(copy.deepcopy(fam))

    def test_grads(self):
        # Test grad function
        grad_diffs = []
        for fam, init_coef, coef_split_idx, ys, Xs in zip(
            self.families, self.init_coefs, self.coef_split_idxs, self.yss, self.Xss
        ):
            grad = fam.gradient(init_coef, coef_split_idx, ys, Xs)

            pos_llk_warp = lambda x: fam.llk(
                x.reshape(-1, 1),
                coef_split_idx,
                ys,
                Xs,
            )
            grad2 = scp.optimize.approx_fprime(init_coef.flatten(), pos_llk_warp)

            grad_diff = np.abs(np.round(grad - grad2.reshape(-1, 1), decimals=3)).max()
            grad_diffs.append(grad_diff)

        grad_diffs = np.array(grad_diffs)
        np.testing.assert_allclose(
            grad_diffs,
            np.array([0.001, 0.001, 0.001, 0.001]),
            atol=min(max_atol, 0.1),
            rtol=min(max_rtol, 0.1),
        )

    def test_total_coef(self):
        total_coefs = []
        for fam, init_coef, coef_split_idx, ys, Xs in zip(
            self.families, self.init_coefs, self.coef_split_idxs, self.yss, self.Xss
        ):
            total_coefs.append(len(init_coef))

        total_coefs = np.array(total_coefs)
        np.testing.assert_allclose(
            total_coefs,
            np.array([39, 21, 39, 21]),
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

    def test_ods(self):
        # Test observation and duration probabilities for hmp
        ods = self.families[0].compute_od_probs(
            self.init_coefs[0],
            self.coef_split_idxs[0],
            self.yss[0],
            self.Xss[0],
            log=False,
        )

        np.testing.assert_allclose(
            np.round(ods[0][0][:5, :, 1], decimals=6),
            np.array(
                [
                    [0.002269, 0.005953, 0.005823, 0.005953, 0.002269],
                    [0.0002, 0.001067, 0.001368, 0.001067, 0.0002],
                    [0.029579, 0.023166, 0.014269, 0.023166, 0.029579],
                    [0.054175, 0.019046, 0.008643, 0.019046, 0.054175],
                    [0.04037, 0.023882, 0.013221, 0.023882, 0.04037],
                ]
            ),
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

        np.testing.assert_allclose(
            ods[0][1][:5, :],
            np.array(
                [
                    [0.10865586, 0.0, 0.10865586, 0.0, 0.10865586],
                    [0.06993352, 0.0, 0.06993352, 0.0, 0.06993352],
                    [0.0539878, 0.0, 0.0539878, 0.0, 0.0539878],
                    [0.04489946, 0.0, 0.04489946, 0.0, 0.04489946],
                    [0.03889596, 1.0, 0.03889596, 1.0, 0.03889596],
                ]
            ),
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

    def test_predict(self):
        # Test signal + state prediction for hmp
        prediction = self.families[2].predict(
            self.init_coefs[2],
            self.coef_split_idxs[2],
            self.yss[2],
            self.Xss[2],
            n_samples=2,
        )

        np.testing.assert_allclose(
            prediction[0][0][:5, :, 0],
            np.array(
                [
                    [1.76405235, 1.62434536, -0.41675785],
                    [1.72963813, 0.35476253, -2.05837901],
                    [2.48217173, 0.92684626, -2.58235028],
                    [2.06388657, 1.97447444, -0.85566706],
                    [0.94255314, 1.26353318, -1.34404703],
                ]
            ),
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

        np.testing.assert_allclose(
            prediction[0][1][:20, 0],
            np.array(
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    2.0,
                    2.0,
                    2.0,
                    2.0,
                    2.0,
                    2.0,
                    2.0,
                    2.0,
                    2.0,
                ]
            ),
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )



import numpy as np
import scipy as scp

# Spline Convolution Code #


def convolve_event(f: np.ndarray, pulse_location: int) -> np.ndarray:
    """Convolution of function ``f`` with dirac delta spike centered around sample
    ``pulse_locations``.

    Based on code by Wierda et al. 2012

    References:
      - Wierda, S. M., van Rijn, H., Taatgen, N. A., & Martens, S. (2012). Pupil dilation \
        deconvolution reveals the dynamics of attention at high temporal resolution. \
        https://doi.org/10.1073/pnas.1201858109

    :param f: Function evaluated over some samples
    :type f: np.ndarray
    :param pulse_location: Location of spike (in sample)
    :type pulse_location: int
    :return: Convolved function as array
    :rtype: np.ndarray
    """
    # Convolution of function f with dirac delta spike centered around
    # sample pulse_locations[i].
    # Based on code by Wierda et al. 2012

    # Create spike
    spike = np.array([0 for _ in range(pulse_location + 1)])
    spike[pulse_location] = 1

    # Convolve "spike" with function template f
    o = scp.signal.fftconvolve(f, spike, mode="full")
    return o


# B-spline functions #


def tpower(x: np.ndarray, t: np.ndarray, p: int) -> np.ndarray:
    """Computes truncated ``p-t`` power function of ``x``.

    Function taken from "Splines, Knots, and Penalties" by Eilers & Marx (2010)

    References:
      - Eilers, P., & Marx, B. (2010). Splines, knots, and penalties. \
        https://doi.org/10.1002/WICS.125

    :param x: Covariate
    :type x: np.ndarray
    :param t: knot location vector
    :type t: np.ndarray
    :param p: degrees of spline basis
    :type p: int
    :return: ``np.power(x - t,p) * (x > t)``
    :rtype: np.ndarray
    """
    return np.power(x - t, p) * (x > t)


def bbase(x: np.ndarray, knots: np.ndarray, dx: float, deg: int) -> np.ndarray:
    """Computes B-spline basis of degree ``deg`` given ``knots`` and interval spacing ``dx``.

    Function taken from "Splines, Knots, and Penalties" by Eilers & Marx (2010)

    References:
      - Eilers, P., & Marx, B. (2010). Splines, knots, and penalties. \
        https://doi.org/10.1002/WICS.125

    :param x: Covariate
    :type x: np.ndarray
    :param knots: knot location vector
    :type knots: np.ndarray
    :param dx: Interval spacing ``(xr-xl) / ndx`` where ``xr`` and ``xl`` are max and min of ``x``
        and ``ndx=nk-deg`` where ``nk`` is the number of basis functions.
    :type dx: float
    :param deg: Degree of basis
    :type deg: int
    :return: numpy.array of shape (-1,``nk``)
    :rtype: np.ndarray
    """
    P = tpower(x[:, None], knots, deg)
    n = P.shape[1]  # Actually n + 1 + 2*deg
    D = np.diff(np.identity(n), n=deg + 1) / (
        scp.special.gamma(deg + 1) * np.power(dx, deg)
    )
    B = np.power(-1, deg + 1) * P @ D
    return B


def B_spline_basis(
    cov: np.ndarray,
    event_onset: int | None,
    nk: int,
    min_c: float | None = None,
    max_c: float | None = None,
    drop_outer_k: bool = False,
    convolve: bool = False,
    deg: int = 3,
) -> np.ndarray:
    """Computes B-spline basis of degree ``deg`` given ``knots``.

    Based on code and definitions in "Splines, Knots, and Penalties" by Eilers & Marx (2010) and
    adapted to allow for convolving B-spline bases.

    References:
      - Eilers, P., & Marx, B. (2010). Splines, knots, and penalties. \
        https://doi.org/10.1002/WICS.125

    :param cov: Flattened covariate array (i.e., of shape (-1,))
    :type cov: np.ndarray
    :param event_onset: Sample on which to place a dirac delta with which the B-spline bases should
        be convolved - ignored if ``convolve==False``.
    :type event_onset: int | None
    :param nk: Number of basis functions to create
    :type nk: int
    :param min_c: Minimum covariate value, defaults to None
    :type min_c: float | None, optional
    :param max_c: Maximum covariate value, defaults to None
    :type max_c: float | None, optional
    :param drop_outer_k: Deprecated, defaults to False
    :type drop_outer_k: bool, optional
    :param convolve: Whether basis functions should be convolved (i.e., time-shifted) with an
        impulse response function triggered at ``event_onset``, defaults to False
    :type convolve: bool, optional
    :param deg: Degree of basis, defaults to 3
    :type deg: int, optional
    :return: An array of shape ``(-1,nk)`` holding the ``nk`` Basis functions evaluated over ``x``
        and optionally convolved with an impulse response function triggered at ``event_onset``
    :rtype: np.ndarray
    """
    # Setup basis with even knot locations.
    # Code based on:
    # Eilers, P., & Marx, B. (2010). Splines, knots, and penalties. https://doi.org/10.1002/WICS.125
    # See also:
    # Eilers, P. H. C., & Marx, B. D. (1996). Flexible smoothing with B-splines and penalties.
    # https://doi.org/10.1214/ss/1038425655

    xl = min(cov)
    xr = max(cov)

    if max_c is not None:
        xr = max_c

    if min_c is not None:
        xl = min_c

    # ndx is equal to n in Eilers & Marx (2011)
    # So there will be n-1 knots (without expansion)
    # n + 1 + 2*deg knots with expansion
    # and nk basis functions computed.
    ndx = nk - deg

    if convolve:
        # For the IR GAMM, the ***outer***-most knots should be close to min_c and max_c.
        # Hence, we simply provide n + 1 + 2*deg (see above) equally spaced knots from
        # min_c to max_c.
        dx = (xr - xl) / (ndx + 2 * deg)
        knots = np.linspace(min_c, max_c, ndx + 1 + 2 * deg)
    else:
        dx = (xr - xl) / ndx
        knots = np.linspace(xl - dx * deg, xr + dx * deg, ndx + 1 + 2 * deg)

    B = bbase(cov, knots, dx, deg)

    if convolve:
        o_restr = np.zeros(B.shape)

        for nki in range(nk):
            o = convolve_event(B[:, nki], event_onset)
            o_restr[:, nki] = o[0 : len(cov)]  # noqa: E203

        B = o_restr

    return B


def TP_basis_calc(cTP: np.ndarray, nB: np.ndarray) -> np.ndarray:
    """Computes row-wise Kroenecker product between ``cTP`` and ``nB``. Useful to create a Tensor
    smooth basis.

    See Wood(2017) 5.6.1 and B.4.

    References:
      - Wood, S. N. (2017). Generalized Additive Models: An Introduction with R, Second Edition \
        (2nd ed.).

    :param cTP: Marginal basis or partially accumulated tensor smooth basis
    :type cTP: np.ndarray
    :param nB: Marginal basis to include in the tensor smooth
    :type nB: np.ndarray
    :return: The row-wise Kroenecker product between ``cTP`` and ``nB``
    :rtype: np.ndarray
    """
    # see: https://docs.scipy.org/doc/scipy/reference/generated/scipy.linalg.khatri_rao.html
    # Function performs col-wise Kron - we need row-wise for the Tensor smooths
    # see Wood(2017) 5.6.1 and B.4
    # ToDo: Sparse calculation might be desirable..
    return scp.linalg.khatri_rao(cTP.T, nB.T).T
